<?php

/**
 * @file
 * Processing (batch/queue) related functionality for Drunkins.
 */

const DRUNKINS_DEFAULT_CRON_TIME = 45;

/**
 * queues operations into a batch set.
 */
function _drunkins_batch_enqueue_items($job_id, array $items, array $job_context, $create_new_batch, $non_progressive = FALSE, $title = '', $return_path = '') {
  /* 'drunkins_rerun' (int) means we're starting a new batch for the Nth time,
   * after finish() has run. In that case, 'drunkins_process_more' is not set
   * and 'drunkins_rerun_start' has no meaning (additional calls to start() are
   * not supported at this point).
   * If 'drunkins_rerun' not set:
   * - 'drunkins_process_more' means another start() call should be made next.
   * - 'drunkins_rerun_start' (int) indicates how many calls have already been
   *    made, unless we're making only one call; then it's unset.
   * - 'drunkins_rerun_start_count' (int) indicates the number of items
   *    already fetched in previous runs.
   *
   * If drunkins_rerun_start > 1**, we will not add a second/third set to the
   * existing batch, but tag the operations onto the same set instead, because:
   * - otherwise we would have to take special measures to carry 'our context'
   *   over to the next batch set (which for us is an arbitrary split of a bunch
   *   of drunkins_batch_process() calls);
   * - working with batch sets isn't logical for us anyway. If we add extra sets
   *   from the drunkins_batch_rerun_start() call (i.e. while running the
   *   first set), the batch_set() will always add an additional set with index
   *   1, and queue the extra items with index 1 - meaning extra batch_set()
   *   calls will add sets in reverse, and all those sets except the last one
   *   will be useless, without items.
   *   (In other words, I think batch_set() is essentially buggy / the
   *   array_slice commands in there are useless, because the queued items are
   *   not moved together with the sets.)
   *
   * (In other words: this function abstracts away the way of queueing
   * operations, but in practice this means that we will execute separate code
   * based on whether we are called from drunkins_batch_rerun_start().)
   *
   * ** In other other words: this means that $create_new_batch will be a
   * parameter to this function, rather than derived. This way, a job cannot
   * mess things up by having drunkins_rerun_start > 1, the first time we pass
   * through here (which will cause fatal errors). When we rewrite this code,
   * the function might be split in two if that turns out better...
   */
//  $create_new_batch = empty($job_context['drunkins_rerun_start'])
//               || $job_context['drunkins_rerun_start'] == 1
//               || !empty($job_context['drunkins_rerun']);
  if ($create_new_batch) {

    // Set up the set with operations. Lose any non-numeric keys for the items.
    // (We could also choose to have $items repeated operations, all calls to
    // drunkins_batch_process() with one item. That will use more memory in
    // this array while queueing but less memory while dequeueing, and it will
    // guard against $items being too big for the queue.data field. But we'll
    // have to do some reorganizing in a.o. drunkins_batch_process() for that?
    // (Discontinue the use of sandbox anywhere?)
    $batch_set = array(
      'operations' => $items ? array(
        array('drunkins_batch_process', array($job_id, array_values($items))),
      ) : array(),
      'file' => drupal_get_path('module', 'drunkins') . '/drunkins.run.inc',
      'finished' => 'drunkins_batch_finished',
      // We can't set sandbox & results here, because these are overwritten by
      // batch_set(). But we can populate other (undocumented) array elements;
      // these will be stored by _batch_populate_queue() and passed to the
      // worker callbacks.
      'job_context' => &$job_context,
    );

    if (empty($title)) {
      $info = drunkins_job_status_info($job_id);
      if (!empty($info['title'])) {
        $batch_set['title'] = $info['title'];
      }
    }
    else {
      $batch_set['title'] = $title;
    }
    if (!empty($job_context['drunkins_rerun'])) {
      $batch_set['init_message'] = t('Re-starting process (@rerun_count) with @count items.',
        array('@rerun_count' => $job_context['drunkins_rerun'], '@count' => count($items)));
    }
    elseif (!empty($job_context['drunkins_rerun_start'])) {
      $batch_set['init_message'] = t('Fetched @count items; trying to fetch more...',
        array('@count' => count($items)));
      // Add function to rerun start() in the same batch set.
      array_unshift($batch_set['operations'],
        array('drunkins_batch_rerun_start', array($job_id)));
    }
    else {
      // The standard progress message is "processing 0 of 1" (because we have 1
      // batch set/step, containing many items. The batch message 'M ::: N' is
      // displayed just below it). Let's make it a bit nicer.
      $batch_set['progress_message'] = t('Processing items...');
    }

    batch_set($batch_set);
    if ($non_progressive) {
      // The definition is set as a 'set' _inside_ a batch. We need to get the
      // whole batch, and set 'progressive' there (in outer definition array).
      $batch = &batch_get();
      $batch['progressive'] = FALSE;
    }
    if ($return_path) {
      // Same for return path. (Won't have effect for non-progressive batches.)
      $batch = &batch_get();
      $batch['source_url'] = $return_path;
    }
  }
  else {
    // We only end up here while executing drunkins_batch_rerun_start().
    $batch = &batch_get();
    if (empty($batch['id']) || count($batch['sets']) != 1
        || !isset($batch['sets'][0]) || isset($batch['sets'][0]['operations'])) {
      _drunkins_log($job_id, t('Internal code error while adding extra items to batch; discarding items!'), array(), WATCHDOG_CRITICAL);
    }
    else {
      // Stick new operation into the existing set and queue it with the same
      // index as the existing ones. It will be executed after the ones already
      // queued (but before the finish callback).
      if ($items) {
        $batch['sets'][0]['operations'] = array(
          array('drunkins_batch_process', array($job_id, array_values($items)))
        );
        _batch_populate_queue($batch, 0);
        // Above is one extra queued operation (there are now multiple
        // 'drunkins_batch_process' operations); account for this.
        $batch['sets'][0]['total']++;
        $batch['sets'][0]['count']++;
        // $job_context is an alias into $batch['sets'][0]['job_context']
      }
    }
    if (!empty($job_context['drunkins_process_more'])) {
      $batch['sets'][0]['progress_message'] = t('Fetched @count items (total @total in @rerun_count iterations); trying to fetch more...',
        array('@rerun_count' => $job_context['drunkins_rerun_start'], '@count' => count($items), '@total' => $job_context['drunkins_rerun_start_count']));
    }
    else {
      // progress_message is not set unless drunkins_process_more was true at
      // one point. However, if it's set, the batch API propagates it to the
      // next step (processing). So overwrite it.
      $batch['sets'][0]['progress_message'] = t('Processing items...');
    }
  }
}

/**
 * 'worker callback' for batch process: get more items.
 *
 * @param string $job_id
 *   The drunkins job ID.
 * @param array|ArrayObject $batch_context
 *   The batch context which the Batch API passes to us. Drupal Core passses
 *   an array, and the Drush code passes an ArrayObject - so we must have it
 *   passed by reference, but cannot use an 'array' typehint.
 */
function drunkins_batch_rerun_start($job_id, &$batch_context) {

  // First time: set up the 'sandbox' and 'results' (persistent data between
  // calls), which we've stored in the batch set's data.
  if (!_drunkins_batch_ensure_context($batch_context)) {
    // This is almost impossible. Abort without specification.
    return;
  }

  $runner_settings = _drunkins_get_process_config($job_id);

  try {
    $items = _drunkins_run_start($job_id, $batch_context['results']);
  }
  catch (\DrunkinsTemporaryError $e) {
    _drunkins_log($job_id, 'Rerunning start for job "@job" threw temporary error; retrying: @error', array('@job' => $job_id, '@error' => $e->getMessage()), WATCHDOG_WARNING);
    // We don't want to wait forever in batch mode; cap at 5 seconds. (This
    // could be configurable if we redo this code.)
    $sleep = min($e->getSuggestedWait(), 5);
    if ($sleep) {
      sleep($sleep);
    }
    $items = array();
  }
  catch (\Exception $e) {
    // The process failed; is already canceled. The queue still contains items
    // from previous fetches though, so we'll need to hack our way around that
    // in drunkins_batch_process(). drunkins_batch_finished() will also be
    // called and we should make sure it does not call the job's finish().
    // Smuggle a flag in the one value that we know is never set here on normal
    // operation:
    $batch_context['results']['drunkins_process_more'] = '*error_on_restart';
    _drunkins_log($job_id, 'Rerunning start for job "@job" threw exception; deleting all previously gathered items from the job. Exception: @error', array('@job' => $job_id, '@error' => $e->getMessage()), WATCHDOG_ERROR);
    return;
  }

  if (!empty($batch_context['results']['drunkins_process_more'])) {
    $batch_context['finished'] = 0;
  }
  if (empty($runner_settings['enqueue_by_fetcher'])) {
    // @todo once we actually implemented 'enqueue_by_fetcher' for batch processes, check whether this if() logic is OK.
    _drunkins_batch_enqueue_items($job_id, $items, $batch_context['results'], FALSE);
  }
}

/**
 * Ensures that our results/sandbox values are set correctly.
 *
 * @param array|ArrayObject $batch_context
 *   The batch context which the Batch API passes to us. Drupal Core passses
 *   an array, and the Drush code passes an ArrayObject - so we must have it
 *   passed by reference, but cannot use an 'array' typehint.
 *
 * @return bool
 *   If our values are indeed set correctly. FALSE indicates fatal error.
 */
function _drunkins_batch_ensure_context(&$batch_context) {

  // 'results' is empty every time we start a new set. We've made sure our
  // 'job_context' always has one value set, so we can recognize if it's filled.
  if (empty($batch_context['results'])) {
    $current_set = &_batch_current_set();
    if (empty($current_set['job_context']) || !is_array($current_set['job_context'])) {
      // Fatal.
      return FALSE;
    }
    // The batch worker code constructed and passed $batch_context for us to
    // use. However, $batch_context['results'] was actually set as a reference
    // to the original location that will get saved. If we reassign something
    // to $batch_context['results'], that reference will get broken if
    // $batch_context is an ArrayObject. (For an array, there's no problem.) We
    // also cannot assign to $batch_context['results'] by reference; that will
    // yield a "Fatal error: Cannot assign by reference to overloaded object"
    // for an ArrayObject. So instead, assign all values inside the 'results'
    // array, to keep the reference intact.
    foreach ($current_set['job_context'] as $key => $value) {
      $batch_context['results'][$key] = $value;
    }
  }

  // 'sandbox' is empty every time we call a new operation.
  if (empty($batch_context['sandbox'])) {
    // One call to this function will process a fixed number of rows:
    // @todo check / document whether this should be a setting - bigger number makes for less calls by Core JS 'batch runner code'?
    if (!isset($batch_context['sandbox']['limit'])
        || !is_numeric($batch_context['sandbox']['limit'])) {
      $batch_context['sandbox']['limit'] = 5;
    }
    $batch_context['sandbox']['progress'] = 0;
  }
  return TRUE;
}

/**
 * 'worker callback' (several items) for batch process: process items.
 *
 * @param string $job_id
 *   The drunkins job ID.
 * @param array $items
 *   Items to process. These are all items for the whole job because we don't
 *   'batch these up' by setting multiple calls to this function in the batch.
 *   (We can change that later if needed.)
 * @param array|ArrayObject $batch_context
 *   The batch context which the Batch API passes to us. Drupal Core passses
 *   an array, and the Drush code passes an ArrayObject - so we must have it
 *   passed by reference, but cannot use an 'array' typehint.
 *
 * @throws \Exception
 *   if the job's processItem() throws an exception and does not have the
 *   runner setting 'handle_process_exception' = 'discard'.
 */
function drunkins_batch_process($job_id, array $items, &$batch_context) {
  if (!$items) {
    // Exit immediately and do not return here.
    return;
  }

  $job = drunkins_get_job($job_id);
  if (!isset($job)) {
    // Should we throw an error here? Which? This is very unlikely to happen.
    return;
  }
  // First time: set up the 'sandbox' and 'results' (persistent data between
  // calls), which we've stored in the batch set's data.
  if (!_drunkins_batch_ensure_context($batch_context)) {
    // This is almost impossible. Abort without specification.
    return;
  }

  if (!empty($batch_context['results']['drunkins_process_more'])) {
    if ($batch_context['results']['drunkins_process_more'] !== '*error_on_restart') {
      // In theory, jobs can do this and we could ignore it. But given that we
      // have at least one 'secret hack' for this value now, we won't.
      throw new LogicException("'drunkins_process_more' context value set during batch processing of items; this should not happen.");
    }
    // Ignore (dequeue without processing) any items that were already queued.
    return;
  }


  $runner_settings = _drunkins_get_process_config($job_id);
  if (!empty($runner_settings['memory_limit'])) {
    _drunkins_set_memory_limit($job_id, $runner_settings['memory_limit']);
  }

  $start_time = time();
  // $count is the number of items in the current batch. 'progress' is the
  // first unprocessed item in the current batch. If we ran start() multiple
  // times, we have multiple batches and we don't know much about the other
  // batches except we likely have 'drunkins_batch_processed_count'.
  $count = count($items);
  if (empty($batch_context['sandbox']['limit']) || !is_numeric($batch_context['sandbox']['limit'])
      || $batch_context['sandbox']['limit'] <= 0) {
    $batch_context['sandbox']['limit'] = $count;
  }
  $limit = min($batch_context['sandbox']['progress'] + $batch_context['sandbox']['limit'], $count);
  $i = $batch_context['sandbox']['progress'];
  while ($i < $limit) {
    $batch_context['results']['drunkins_progress'] = ++$i;
    if (!empty($batch_context['results']['drunkins_rerun_start_count'])) {
      // Get a progress counter that increases over batches.
      if (empty($batch_context['results']['drunkins_batch_processed_count'])) {
        $batch_context['results']['drunkins_batch_processed_count'] = 0;
      }
      $batch_context['results']['drunkins_batch_processed_count']++;
    }

    try {
      $job->processItem($items[$i - 1], $batch_context['results']);
    }
    catch (\Exception $e) {
      $item = $items[$i - 1];
      // I _think_ the batch runner wants to honour the *_log_* settings just
      // like the queue runner, though this can make the messages pretty big.
      // Tweak as needed. Below is an almost-straight copy of
      // _drunkins_cron_process_queue_item() code, but without the comments and
      // with $job_id / $item changed:
      $handle = $e instanceof \DrunkinsTemporaryError && isset($runner_settings['handle_process_temp_error'])
        ? $runner_settings['handle_process_temp_error']
        : (isset($runner_settings['handle_process_exception']) ? $runner_settings['handle_process_exception']
          : 'fallthrough');
      if ($handle !== 'fallthrough') {
        $log_message = $e instanceof \DrunkinsTemporaryError ? 'encountered a temporary error' : 'failed';
        if ($handle === 'discard') {
          $log_message .= '; discarding item';
        }
        $log_message = "Processing '@job' item $log_message: @message";
        $log_params = ['@job' => $job_id, '@message' => $e->getMessage()];
        $log_item = $e instanceof \DrunkinsTemporaryError && isset($runner_settings['process_temp_error_log_item'])
          ? $runner_settings['process_temp_error_log_item']
          : (isset($runner_settings['process_exception_log_item']) ? $runner_settings['process_exception_log_item']
            : FALSE);
        if ($log_item) {
          $json_pretty_format = stripos($log_item, 'json') === 0 ? (bool) stripos($log_item, 'pretty') : NULL;
          $log_message .= $json_pretty_format === FALSE ? '<br>Item: @item' : '<br>Item: <pre>@item</pre>';
          $log_params['@item'] = $json_pretty_format === NULL
            ? var_export($item, TRUE) : json_encode($item, $json_pretty_format ? JSON_PRETTY_PRINT : 0);
        }
        if ($e instanceof \DrunkinsTemporaryError && isset($runner_settings['process_temp_error_log_trace'])
          ? $runner_settings['process_temp_error_log_trace']
          : (isset($runner_settings['process_exception_log_trace']) ? $runner_settings['process_exception_log_trace']
            : FALSE)) {
          $log_message .= '<br>Trace: <pre>@trace</pre>';
          $log_params['@trace'] = $e->getTraceAsString();
        }
      }
      if ($handle === 'discard') {
        _drunkins_log($job_id, $log_message, $log_params, WATCHDOG_ERROR);
        // Do accounting, and wait if the exception indicates it. The item will
        // be dequeued.
        if ($e instanceof \DrunkinsTemporaryError) {
          $batch_context['results']['drunkins_temp_error_count']++;
          // Wait for the suggested wait time, capped to the maximum time we
          // think it is safe to wait while not letting this HTTP request time
          // out.
          // @todo this should be a setting, or be derived otherwise. For now,
          //   we'll be conservative (i.e. we might wait less than suggested)
          //   and have a higher chance of encountering an error while
          //   processing the next item as well. There are more important TODOs
          //   to figure out before this one.
          $assumed_max_http_request_time = 30;
          $already_ran_for = time() - $start_time;
          // $assumed_max_http_request_time clearly has to be larger than
          // $already_ran_for, though. If it's not, then make it the smallest
          // multiple of itself that is larger than $already_ran_for - which
          // has the same result as subtracting $assumed_max_http_request_time
          // from $already_ran_for until it's smaller.
          $sleep = min($assumed_max_http_request_time - ($already_ran_for % $assumed_max_http_request_time), $e->getSuggestedWait());
          if ($sleep) {
            sleep($sleep);
          }
        }
        else {
          // A regular exception has no 'suggested wait time' here and we don't
          // by default assume that there is anything time based about whatever
          // caused the exception to be thrown. So continue immediately.
          $batch_context['results']['drunkins_exception_count']++;
        }
      }
      else {
        // 'log' / 'fallthrough'. (Other values should have been caught before
        // process start.) For batch, this usually means the caller will
        // display the exception as an error message, in a screen with a
        // "Please continue to the error page." link. (The target page will
        // run the 'finished' callback which will not run finish() but reset the
        // process state - and display another message about an exception).
        if ($handle === 'log') {
          // We have to log more than just the message, based on the above log
          // settings. I _think_ that also goes for the batch runner.
          _drunkins_log($job_id, $log_message, $log_params, WATCHDOG_ERROR);
        }
        throw $e;
      }
    }
  }
  $batch_context['sandbox']['progress'] = $i;

  // Set status message and progress indicator, whether or not we've finished.
  // This function will be called again when $batch_context['finished']
  // is explicitly set to a value < 1. (It's initalized to 1 by the caller.)
  if ($count == 0) {
    $batch_context['message'] = 'Aborted.';
  }
  else {
    $batch_context['finished'] = $i / $count;
    $batch_context['message'] = $i . ' ::: ' . $count;
    if (!empty($batch_context['results']['drunkins_rerun_start_count'])) {
      // We apparently have multiple batches and $count is only one batch; make
      // an attempt to also display the full total count being processed. The
      // 'current batch count' that is already in the message doesn't mean much
      // but we'll append to it instead of replacing it... to see if we can
      // indeed trust this. (We haven't really tested well whether it works.)
      $batch_context['message'] .= ' / ' . $batch_context['results']['drunkins_batch_processed_count']
                                   . ' ::: ' . $batch_context['results']['drunkins_rerun_start_count'];
    }
  }

  // This one is necessary for the 'finished callback'; set it every time to
  // protect against messing by processItem().
  $batch_context['results']['drunkins_name'] = $job_id;
}

/**
 * 'finished callback' for batch process.
 */
function drunkins_batch_finished($success, array $job_context, $operations) {
  if (!empty($job_context['drunkins_name'])) {
    $job_id = $job_context['drunkins_name'];
  }
  // Sometimes when a fatal error has happened, $job_context (in other words,
  // batch_set['results']) is empty. Check if $operations still contains the
  // 'drunkins_batch_process' call with items. If so, we can get the job ID from
  // there.
  elseif (isset($operations[0][0]) && $operations[0][0] == 'drunkins_batch_process'
          && isset($operations[0][1][0])) {
    $job_id = $operations[0][1][0];
  }

  // We just added this at some point to be able to do the below check but
  // really hope to be able to just get rid of this completely / re-do it /
  // decide that we never have a null $job_id...
  if (!isset($job_id) && array_filter(drupal_static('_drunkins_is_running_non_progressive'))) {
    throw new LogicException('job id not set in drunkins_batch_finished(), while a job is running non-progressively. We cannot decide what to do.');
  }
  $non_progressive = _drunkins_is_running_non_progressive($job_id);

  $message = '';
  $non_progressive_message = '';

  // The 'success' parameter means no fatal PHP errors were detected. All
  // other error management should be handled using $job_context.
  if ($success) {
    if (isset($job_context['drunkins_process_more'])
        && $job_context['drunkins_process_more'] === '*error_on_restart') {
      // _drunkins_run_start actually logged a message so this is duplicate.
      // Hack in a message for the non-progressive case which demands a return
      // value at the moment; this message won't be used anywhere for the
      // regular progressive case.
      // @todo reevaluate after rewrite of return values?
      $non_progressive_message = t('Starting job "@job" was done in multiple phases; one of these phases failed.', array('@job' => $job_id));
    }
  }
  if (!$non_progressive_message) {
    if (!isset($job_id)) {
      _drunkins_log('?', "Batch 'finished callback' could not derive job id, so could not properly finish the job.", array(), WATCHDOG_ERROR);
      $job_id = '?';
    }
    if (!$success) {
      _drunkins_log($job_id, 'Job "@job" encountered a fatal error and did not properly finish.', array('@job' => $job_id), WATCHDOG_ERROR);
    }
    elseif ($job_id !== '?') {
      // Clear out the 'continue' value, to make sure we'll only use it when
      // finish() explicitly sets it. (Theoretically a processItem() could set
      // it, though we don't support this yet; we'll protect against it and
      // won't warn.)
  // @todo consider if we want to make this array (part of) the return
  // argument, instead of the message (if we want to get rid of that anyway)
      unset($job_context['drunkins_process_more']);
      $job = drunkins_get_job($job_id);
      if (!$job) {
        _drunkins_log($job_id, "Batch 'finished callback' failed to create job \"@job\"", array('@job' => $job_id), WATCHDOG_ERROR);
      }
      else {
        // Log before finish() (at debug level).
        // @todo what do we exactly want to do immediately after, when finish() does not return a message anymore?
        _drunkins_log($job_id, 'Calling finish() for job "@job".', array('@job' => $job_id), WATCHDOG_DEBUG);
        $runner_settings = _drunkins_get_process_config($job_id);
        if (!empty($runner_settings['memory_limit_finish']) || !empty($runner_settings['memory_limit'])) {
          _drunkins_set_memory_limit($job_id, !empty($runner_settings['memory_limit_finish'])
            ? $runner_settings['memory_limit_finish'] : $runner_settings['memory_limit']);
        }
        try {
          $message = $job->finish($job_context);
        }
        catch (\Exception $e) {
          // @todo ideally, we should probably also have some special handling
          //   for DrunkinsTemporaryError.
          // @todo if we make the items the return value, we can simplify a lot below because we cannot generate an exception AND more items.
          //  We'll also have to think about the other @todo below though.
          if (isset($runner_settings['handle_finish_exception'])
              && ($runner_settings['handle_finish_exception'] === 'continue'
                  || $runner_settings['handle_finish_exception'] === 'finish')) {
            // Log exception and finish normally (or continue on more items).
            _drunkins_log($job_id, 'Calling finish() for job "@job" failed: @message', array('@job' => $job_id, '@message' => $e->getMessage()), WATCHDOG_ERROR);
            if (!empty($job_context['drunkins_process_more']) && is_array($job_context['drunkins_process_more'])) {
              if ($runner_settings['continue_on_finish_exception'] === 'continue') {
                _drunkins_log($job_id, '"@job" finish() generated @count more items to process: continuing despite exception thrown.', array('@job' => $job_id), WATCHDOG_NOTICE);
              }
              else {
                _drunkins_log($job_id, '"@job" finish() generated @count more items to process: discarding them because of exception thrown.', array('@job' => $job_id), WATCHDOG_NOTICE);
                unset($job_context['drunkins_process_more']);
              }
            }
          }
          else {
            // Default value: 'break'. (Other values should have been caught
            // before process start.) We now have two options:
            // - continue without running _drunkins_finish_process() (The
            //   standard UI will redirect to the start screen again, where the
            //   job cannot be started again; the current process will be stuck
            //   with 0 items to process, until a user manually resets it.
            // - throw the exception, which will have Drupal generate a plain
            //   "unexpected error; please try again" screen without any links
            //   to e.g. finish the batch process. (The URL still includes
            //   '&op=finished' so releasing the page will indeed run finish()
            //   again. If the exception persists, the user must explicitly
            //   navigate away from the page and the current process is still
            //   stuck like it would be for the first option.)
            // We choose the latter because it's more 'in your face'.
            // (Note if we ever get an option to run finish() for batch
            // processes with 0 items like we do for queue processes... this
            // whole thing needs to be reevaluated. See also
            // _drunkins_queue_process_check_finish() for that case.)

            // If a process returned more items to process, we'll always need to
            // log that fact here, despite not always logging the exception itself
            // here (if the caller takes care of it). This means the "discarding
            // items" log will happen a bit out of order.


            // We know (or assume) the caller will log the exception. We don't
            // know what function it will log as the source though, so log the
            // fact that the exception was thrown in finish().
            _drunkins_log($job_id, 'Calling finish() for job "@job" generated an exception.', array('@job' => $job_id), WATCHDOG_NOTICE);
            // We're not logging about any extra items; they *probably* will be
            // saved into context (if the PHP 'shutdown functions' still run
            // after an exception; haven't tested that) and be processed if the
            // user reloads and triggers finish() and that finishes without
            // throwing an exception next time.
            // @todo this completely changes if we make the extra items a return
            //   value instead of a subkey in $context. Think about this.

            throw $e;
          }
        }
      }
    }
  }

  // From here on, job ID will mostly be used for logging.
  if (!isset($job_id)) {
    $job_id = '?';
  }
  if (!empty($job_context['drunkins_process_more']) && is_array($job_context['drunkins_process_more'])) {
    // If we got any message back, don't just drop it; log/print it. (Print
    // by default, unlike everywhere else.)
    if ($message) {
      _drunkins_log($job_id, $message, array(), WATCHDOG_INFO);
    }
// @todo maybe make the 'executive decision' that status logs like the below
// will not be done by the runner, and logging is the exclusive responsibility
// of the job (except in clear errors)? Not sure; we should treat it the same as the 'N items received' from start().
    _drunkins_log($job_id,
      'finish() for job "@job" returned @count new items; continuing...',
      array('@job' => $job_id, '@count' => count($job_context['drunkins_process_more'])),
      WATCHDOG_INFO);

    $batch = &batch_get();
    // We're being called from _batch_finished(). That function processes things
    // differently depending on whether the batch is progressive / processed
    // from a web UI form:
    if (!$non_progressive && !drupal_is_cli()) {
      // Remember context because our caller will clear out the batch variables.
      $continue_context = &drupal_static('drunkins_batch_continue', array());
      $continue_context = $job_context;

      // Intercept the caller by making it pass through drupal_redirect_form()
      // (which is always called even though we are on a regular page, not a
      // form) and to our own callback.
      $batch['form_state']['no_redirect'] = TRUE;
      $batch['redirect_callback'] = 'drunkins_batch_continue';
      // We need to remember the source URL to redirect to after continuing, and
      // $batch['url'] does not have any function anymore now that we don't
      // redirect automatically. So remember it in $batch['url']; then it will
      // get passed to the redirect callback.
      $batch['url'] = $batch['source_url'];
    }
    else {
      // _batch_finished() will just exit for non-progressive batches (and
      // _drush_batch_finished() will exit for all types of batches) so we
      // don't need to do funny overriding, and have to continue by ourselves.
      // This does mean we're part of a recursive call chain now, though: after
      // we're done processing the new items and have executed this function
      // recursively, drunkins_batch_continue() will return here and we'll
      // return to our caller.
      $return_path = drupal_is_cli() ? '' : $batch['source_url'];
      drunkins_batch_continue($return_path, $job_context);
    }
  }
  else {
    // Log summary message about processItem() throwing exceptions. (Each has
    // logged an individual error but it seems good to do a summary - with
    // WARNING severity then.)
    if (!empty($job_context['drunkins_exception_count'])) {
      _drunkins_log($job_id, '@count exceptions were thrown while processing items for job "@job".',
        array('@job' => $job_id, '@count' => $job_context['drunkins_exception_count']), WATCHDOG_WARNING);
    }

    if ($non_progressive) {
      // Reset memory limit to initial one. Do it here so it's done only once if
      // we were called recursively through drunkins_batch_continue(). For
      // progressive batches we don't care because the caller will redirect.
      _drunkins_set_memory_limit($job_id);

      // Since no batch operations/finish callback return anything to the
      // caller, pass the context back through a static variable.
      $log = &drupal_static('drunkins_last_log');
      $log = array($non_progressive_message ? $non_progressive_message : $message, array(), WATCHDOG_INFO, $job_context);

      _drunkins_set_running_non_progressive($job_id, FALSE);
    }
    // For 'normal' jobs, report message here. If the process did not return a
    // message, that's fine.
    elseif (!empty($message)) {
      drupal_set_message($message);
    }

    // In the '*error_on_restart' case we already canceled the process. (We just
    // get here because we couldn't delete process/finish callbacks anymore.)
    if ((!isset($job_context['drunkins_process_more'])
         || $job_context['drunkins_process_more'] !== '*error_on_restart')
        && $job_id !== '?'
    ) {
      _drunkins_finish_process($job_id, array('delete_items' => FALSE), $job_context);
    }
  }
}

/**
 * Start up a new batch process with items received from a previous batch'
 * 'finished' operation.
 *
 * @todo implement automated tests. the latest changes have never been tested,
 *   though they seem harmless.
 */
function drunkins_batch_continue($path, array $job_context) {
  $progressive_noncli_context = &drupal_static('drunkins_batch_continue', array());
  // We assume that if this is nonempty, we're a progressive batch job that
  // doesn't run from drush - and if this is empty, we're either
  // non-progressive or running from Drush. (There's some tie-in with
  // $non_progressive below, but we unfortunately need to store these in two
  // separate storages. For now?
  if ($progressive_noncli_context) {
    // We're called as 'redirect_callback', so we need to get our argument from
    // a stored static variable.
    $job_context = $progressive_noncli_context;
    $progressive_noncli_context = array();

    // The difference between this function and drunkins_batch_start() is, we
    // are not called from a submit function so the grandparent process_form()
    // does not call calls batch_process().
    // Now that we overrode the 'drupal_goto' to get here, our caller will do
    // nothing, return nothing, output no HTML. So we will call batch_process()
    // ourselves, which initializes a new batch with the above set and calls
    // drupal_goto().
    // The batch has been removed from the database table. We don't need it; we
    // can just call batch_set() for a new batch, and batch_process() will fill
    // the database table and start things as a new process with incremented ID.
  }
  else {
    // The finished batch still exists. (As opposed to a progressive batch
    // using Core's worker, above: _batch_finished() has deleted the finished
    // batch and calling batch_set() will create a new batch.) We need to
    // delete it; we can't add another set to the existing batch, because
    // _batch_finished() would end up calling the 'finished callback' for the
    // first set again later.
    $batch = &batch_get();
    $batch = NULL;

    // Since we don't do funny overrides or redirects, this creates a recursive
    // call tree: batch_process() -> _batch_process() -> _batch_finished() ->
    // drunkins_batch_finished() -> drunkins_batch_continue() ->
    // batch_process(). (When running through the Drush batch worker, this also
    // applies, except drush_backend_batch_process() spawns a new process and
    // we wait for it to finish.) All these functions will return immediately
    // after drunkins_batch_finished() yields no items to process... except
    // _batch_finished(), which tries to do some $_SESSION cleanup. We could try
    // to do proper $_SESSION cleanup because we're changing the batch ID before
    // executing that code... but it's buggy anyway, so it's clearly not used.
  }

  $non_progressive = _drunkins_is_running_non_progressive($job_context['drunkins_name']);

  // Set a new batch.
  // Unset drunkins_process_more because we don't want it stored.
  $items = $job_context['drunkins_process_more'];
  unset($job_context['drunkins_process_more']);
  if (isset($job_context['drunkins_rerun']) && is_numeric($job_context['drunkins_rerun'])) {
    $job_context['drunkins_rerun']++;
  }
  else {
    $job_context['drunkins_rerun'] = 1;
  }
  _drunkins_batch_enqueue_items($job_context['drunkins_name'], $items, $job_context, TRUE, $non_progressive,
    !empty($job_context['drunkins_process_more_title']) ? $job_context['drunkins_process_more_title'] : '',
    $path
  );
  unset($items);

  if ($non_progressive) {
    // Reset memory limit from the 'finish' one, before doing more processing
    // but after queueing the items. For progressive batches we're not in the
    // same HTTP request anymore as when calling finish().
    $runner_settings = _drunkins_get_process_config($job_context['drunkins_name']);
    if (!empty($runner_settings['memory_limit_finish'])) {
      if (!empty($runner_settings['memory_limit'])) {
        $initial_limit = _drunkins_get_initial_memory_limit();
        _drunkins_set_memory_limit($job_context['drunkins_name'], $initial_limit > $runner_settings['memory_limit'] ? $initial_limit : $runner_settings['memory_limit'], TRUE);
      }
      else {
        _drunkins_set_memory_limit($job_context['drunkins_name']);
      }
    }
  }

  (!$non_progressive && drupal_is_cli()) ? drush_backend_batch_process() : batch_process();
}

/**
 * Claims a job process, runs the job's start() method and initializes all
 * related job info (context, run state, database table, etc). This function
 * abstracts away the icky details around keeping the cached job info and
 * database table in sync while the job process' state changes.
 *
 * It's meant to only be called from drunkins_start_job(); it's only separated
 * to isolate these icky details and because drunkins_start_job() gets too long.
 *
 * @param string $job_id
 * @param string $method
 * @param array $context
 *   Initial context (usually an empty array) passed by reference if the caller
 *   needs it. (The reference is usually for batch processes because this
 *   function stores the context for queue processes but not for batch.)
 *
 * @return array
 *   Items as returned by the job's start() method.
 *
 * @throws \Exception
 *   Any exception or own error condition encountered. This function will
 *   already log the error so the caller will not have to do much with it (which
 *   is good because this function is an integrated part of the caller).
 *
 * @see drunkins_start_job()
 */
function _drunkins_start_process($job_id, $method, array &$context) {
  // Get $settings and $runner_settings into variables just because we need to
  // store them twice in a related array. (If we start storing them in separate
  // database fields, the code may change.)
  $runner_settings = _drunkins_get_process_config($job_id);
  $settings = _drunkins_get_process_config($job_id, 'job');

  if ($method === 'batch') {
    $init_data = array('settings' => $settings, 'runner_settings' => $runner_settings);
    // For batch runs, we don't store 'context' in the drunkins_process table;
    // it will be stored somewhere else by the caller. (Likely the batch table,
    // or a MemoryQueue for non-progressive batches through batch_set())
  }
  else {
    // drunkins_progress for now is a kind-of-half-officially-documented
    // progress indicator; it's 1-based (1 for the first processed item). For
    // batch processes it just copies the 'sandbox][progress' value but for
    // queue processes we need to initialize it here.
    // We may revisit and/or document this better, once we figure out what to
    // do with concurrency of processing items (which is not supported - but we
    // mostly ignore it until now and will properly document things later).
    // @todo make docs about concurrency of processing items etc
    $context['drunkins_progress'] = 0;
    $init_data = array(
      'settings' => $settings,
      'context' => $context,
      'runner_settings' => $runner_settings
    );
  }
  // Another not-officially-documented value: the start date. Until now only
  // used in _drunkins_run_start() and not guaranteed to continue existing.
  $context['drunkins_process_start'] = time();
  $init_fields = array(
    'method' => $method,
    'run_info' => 1,
    'started' => $context['drunkins_process_start'],
  );
  $calc_init_fields = array(
    'running' => 3,
  );

  // Gather fields for DB update.
  $db_fields = $init_fields + array('data' => serialize($init_data));
  // Try to update the run status of the process in a 'safe' way: only one
  // thread can succeed in UPDATEing the same row.
  $status = db_update('drunkins_process')
    ->fields($db_fields)
    ->condition('queue_name', $job_id)
    ->where('((run_info = 0 AND finished >= started) OR run_info > 3)')
    ->execute();
  if (!$status) {
    // We didn't succeed in updating a row; maybe there is none for $job_id yet.
    // Try to insert one. If another PHP process just claimed the job, this
    // will throw a database exception. Since our table has no serial ID, the
    // return value of the insert query is always 0 - so we will need to check &
    // lock the database row with a repeated update query.
    db_insert('drunkins_process')
      ->fields(array(
        'queue_name' => $job_id,
        'method'     => $method,
        'run_info'   => 0,
        'started'    => 0,
        'finished'   => 0,
        'data'       => serialize(array()),
      ))
      ->execute();
    $status = db_update('drunkins_process')
      ->fields($db_fields)
      ->condition('queue_name', $job_id)
      ->where('((run_info = 0 AND finished >= started) OR run_info > 3)')
      ->execute();
  }
  if (!$status) {
    throw new \Exception('Failed to claim job; drunkins_process row update failed somehow. This is strange.');
  }
  // NOTE: after this point, batch processes must not call
  // _drunkins_set_process_config() anymore because the changed config will not
  // be saved in the database automatically. (There is at least one location
  // which does this, but does a db_update call afterwards. We hope we can
  // reorganize the code sometime.)

  // Update the cached job status info to match the database state.
  // (Otherwise in the remainder of this HTTP request, drunkins_get_job()
  // will get the wrong settings, drunkins_cron_process_queue_item() will
  // get no context and discard items, etc.) This is not very important (we
  // probably won't use the data) until the job actually starts running, but we
  // try to be as complete/early as possible to prevent obscure bugs. (Also,
  // now that we do that anyway, we'll use the info in _drunkins_run_start().)
  $info = &drunkins_job_status_info($job_id);
  // Set 'running' already; necessary to make both _drunkins_run_start() and
  // _drunkins_finish_process() behave correctly.
  $info = array_merge($info, $init_fields, $calc_init_fields, array(
    'process_settings' => $settings,
    // Set context by reference so $context is modified after exiting this
    // function, but use $info['process_context'] below. To be changed when
    // this function is retired.
    'process_context' => &$context,
    'process_runner_settings' => $runner_settings,
  ));

  // Unset these context values on first run, to prevent strange conditions.
  // Jobs should not change these.
  unset($info['process_context']['drunkins_rerun_start']);
  unset($info['process_context']['drunkins_rerun_start_count']);

  $items = _drunkins_run_start($job_id, $info['process_context']);

  // A second kind-of-half-officially-documented indicator for processing runs:
  $info['process_context']['drunkins_exception_count'] = 0;

  $items_queued = empty($runner_settings['enqueue_by_fetcher']) ? count($items) : DrupalQueue::get($job_id)->numberOfItems();
  if (!$items_queued && empty($info['process_context']['drunkins_process_more'])) {
    _drunkins_log($job_id, 'There are no items to process for job "@job".', array('@job' => $job_id),
      isset($runner_settings['log_severity_no_items']) ? $runner_settings['log_severity_no_items'] : WATCHDOG_INFO);
    _drunkins_finish_process($job_id, array('target_state' => 'empty_start', 'delete_items' => FALSE), $info['process_context']);
  }
  elseif ($method !== 'batch') {
    // Also if start() reset $context, we still want to set 'progress'.
    $info['process_context']['drunkins_progress'] = 0;
    if ($info['process_context'] !== $init_data['context']) {
      // Context has been changed by start(), so it must be updated in the
      // database. We'll postpone this if possible, because the context may
      // still be used & changed by drunkins_cron_process_queue_item(). Call the
      // 'current context' function with this job ID to mark its context as
      // needing to be saved. (This registers a shutdown function. If we started
      // another job in the same HTTP request, that job's context will be
      // written to the database now. So if we initialize job A and job B, and
      // then process queued items for job A and B, in one HTTP request... we
      // make two extra database calls for writing context, which could be
      // optimized away. But that's not worth the effort.)
      drunkins_current_context($job_id);
    }
  }

  return $items;
}

/**
 * Runs start() for a job; handles some errors.
 *
 * This also sets drunkins_process_more/_rerun_start values in $job_context.
 * Normally, no logging is done about number of items returned/need to continue;
 * that's up to the caller (and should be done when items are actually queued).
 * This function only logs before calling start (severity: debug), logs that
 * items are queued if 'enqueue_by_fetcher' is set, and logs (most, not all)
 * 'fatal' errors. In error cases, the process info (in context and db) is reset
 * if this is NOT the first start() run; if it is the first, the caller is
 * responsible for that.
 * @todo revisit above once we make this into a runner class? At the moment, the
 *  full info for saying 'status is batch-error-on-start vs queue-error-on-start
 *  is not known to us. Hence the inconsistency in process info resetting.
 *
 * Besides the input arguments, this function also uses (a subset of)
 * process_runner_settings in status info.
 *
 * This function changes 'drunkins_rerun_start(_count)' values in $job_context.
 * The 'drunkins_process_more' value which can be set by the job class, is used
 * as input for this. (Also, it always unsets 'drunkins_override_fetch'.)
 *
 * @param $job_id
 *   Job ID / machine name for the the queue.
 * @param array $job_context
 *   Context, which will be altered by the job and this function. (Note that
 *   even though other things will be altered directly into the job info, we
 *   must use this argument, not the job info, to alter context.)
 *
 * @return array
 *   Items as returned by the job's start() method.
 *
 * @throws \Exception
 *   Any exception or own error condition encountered. This function will
 *   already log the error. Process info (in context and db) is reset if this is
 *   NOT the first start() run; if it is, the caller is responsible for that.
 * @todo revisit above once we make this into a runner class? At the moment, the
 *  full info for saying 'status is batch-error-on-start vs queue-error-on-start
 *  is not known to us. Hence the inconsistency in process info resetting.
 */
function _drunkins_run_start($job_id, array &$job_context) {

  // Remember if this is the very first run in the process (which means:
  // - the process never indicated 'process_more' - from start(), processItem()
  //   or finish();
  // - the caller is _drunkins_start_process().
  // (If these 2 things are not equivalent, this comment should be rewritten :))
  $first_run = empty($job_context['drunkins_rerun_start']);
  // Note the job class itself is not meant to set 'drunkins_rerun_start' and
  // '..._count' values but it can use them; it means how many times start() had
  // already been called / how many items had been returned in _previous_ calls.
  // If no calls have been made, both are unset.

  $job = drunkins_get_job($job_id);
  if (!isset($job)) {
    // This will happen e.g. if the 'class' was not defined.
    throw new \Exception( 'Failed to create job class instance.');
  }

  // If we have custom items, pass them to start() in context.
  $items = _drunkins_get_process_config($job_id, 'items');
  if ($items !== FALSE) {
    $job_context['drunkins_override_fetch'] = $items;
    // We don't know how many items these can be, so for memory management,
    // unset these.
    _drunkins_set_process_config($job_id, 'items', array());
  }

  unset($job_context['drunkins_process_more']);

  // Log before start() (at debug level). Logging immediately after start() is
  // not done explicitly, unless 'enqueue_by_fetcher' is set: we usually wait
  // until items are queued.
  _drunkins_log($job_id, 'Calling start() for job "@job".', array('@job' => $job_id), WATCHDOG_DEBUG);
  // 'process_settings' and 'process_runner_settings' are set. (If this is the
  // first run: the process info has been updated to match the process as if it
  // was already running, just before calling us.)
  $runner_settings = _drunkins_get_process_config($job_id);
  // Set memory limit. We will not reset it; this is left up to the caller after
  // items are queued.
  if (!empty($runner_settings['memory_limit_start']) || !empty($runner_settings['memory_limit'])) {
    _drunkins_set_memory_limit($job_id, !empty($runner_settings['memory_limit_start'])
      ? $runner_settings['memory_limit_start'] : $runner_settings['memory_limit']);
  }
  try {
    $items = $job->start($job_context);
    // Protect against hacks/errors in the job class.
    if (!is_array($job_context)) {
      // We haven't defined our own exception type yet. Maybe later, if there's
      // a need.
      throw new \Exception('The context has a wrong format after returning from start().');
    }
  }
  catch (\Exception $e) {
    // Abort the process unless this is a DrunkinsTemporaryError AND we have
    // called start() for the not-first time. (We always abort after a first
    // start() call, for now; not doing so requires more changes to callers.)
    $continue = $e instanceof DrunkinsTemporaryError && !$first_run;
    if ($continue) {
      $job_context['drunkins_process_more'] = TRUE;
      $job_context['drunkins_rerun_start']++;
    }
    else {
      _drunkins_finish_process($job_id, $first_run
        ? array('target_state' => 'error_on_start', 'delete_items' => FALSE)
        : array('target_state' => 'error_on_restart'),
        is_array($job_context) ? $job_context : []);
    }
    throw $e;
  }

  // Remove items if they were present because we don't want to keep / store
  // them in process state. (Only necessary for non-batch mode.) We lose them if
  // there's an error between this point and queuing $items - but that isn't
  // worse than losing normal $items.
  unset($job_context['drunkins_override_fetch']);

  if (variable_get('drunkins_save_process_summary', FALSE)) {
    if (is_callable(array($job, 'initializeSimpleCounters'))) {
      // Remember the number of items we're going to enqueue, regardless whether
      // we'll save them later (though saving them later is the only currently
      // known application. We may want to deduplicate this with
      // 'drunkins_rerun_start_count' later, but for now we'll just live with
      // duplicate counts because of slightly different applications - and my
      // unwillingness to accommodate 'enqueue_by_fetcher' for this number now.)
      // We're naming the counter 'queued' even though we increase it before the
      // items are actually queued, for simplicitly. This first call is a little
      // bit wasteful when start was re-run but it's OK...
      $job->initializeSimpleCounters(['drunkins_queued'], FALSE, $job_context);
      $job->increaseCounter('drunkins_queued', count($items), $job_context);
    }
    // If we're going to save statistics, we insert a row with process ID and
    // start date. We'll use ID as an indicator that the row is already there
    // if we're re-running start(). adding the finish time is done in
    // _drunkins_finish_process(). If there are no items to process so we
    // immediately proceed to finishing off the process, that means we'll make
    // two database queries which we could have done in one, but that's not a
    // big concern right now.
    if (empty($job_context['drunkins_process_id']) && !empty($job_context['drunkins_process_start'])) {
      $job_context['drunkins_process_id'] = db_insert('drunkins_process_run')
        ->fields(['name' => $job_id, 'started' => $job_context['drunkins_process_start']])
        ->execute();
    }
  }

  // Remember the number of times we've run, only if 'drunkins_process_more' was
  // set (now or in the previous run). A job shouldn't mess with the
  // 'drunkins_rerun_start' value, but if it does we reset it to 1.
  if (!empty($job_context['drunkins_rerun_start']) && is_int($job_context['drunkins_rerun_start'])) {
    $job_context['drunkins_rerun_start']++;
    if (empty($runner_settings['enqueue_by_fetcher'])
        && isset($job_context['drunkins_rerun_start_count'])
        && is_numeric($job_context['drunkins_rerun_start_count'])) {
      $job_context['drunkins_rerun_start_count'] += count($items);
    }
    else {
      // If 'enqueue_by_fetcher' or enqueueing extra items explicitly, we don't
      // know if items were dequeued by now.
      $job_context['drunkins_rerun_start_count'] = '?';
    }
  }
  elseif (!empty($job_context['drunkins_process_more'])) {
    $job_context['drunkins_rerun_start'] = 1;
    $job_context['drunkins_rerun_start_count'] = empty($runner_settings['enqueue_by_fetcher'])
      ? count($items) : DrupalQueue::get($job_id)->numberOfItems();
  }

  return $items;
}

/**
 * Enqueue items into a running process.
 *
 * This function is not suited for calling start() before a process is really
 * started because it does not change a process' recorded state to "running".
 * (It only changes process state to "cancelled" if a start() call fails.) So at
 * the first call during a process' lifetime, $items will probably be passed
 * (because the rest of the code is set up such that starting a job also means
 * getting items returned), though the function does not depend on this.
 *
 * If the 'progressive' runner setting is explicitly set to FALSE, start() may
 * be called (repeatedly) during this one function call, to enqueue more items,
 * so the 'start phase' of the process always gets handled in one call.
 *
 * If 'progressive' is true/unset, the caller is expected to handle situations
 * where the job's start() signifies more items should be fetched, i.e. to call
 * this function repeatedly. Since the process status then is already "running",
 * the $items argument may be left empty so that start() is run by this function
 * before enqueueing its resulting items.
 *
 * If a start() call fails, the whole process will be canceled (as opposed to
 * just skipping the result of this one call).
 *
 * @param string $job_id
 *   Job ID, also used as machine name of the queue.
 * @param array $items
 *   (optional) $items to enqueue. If passed, the job's start() method will be
 *   run to fetch them, before queueing them.
 * @param array $override_runner_settings
 *   (optional) A specific subset of runner settings which can override the
 *   running process' settings for this call: 'logging', 'log_level' and
 *   'add_last_marker'. 'add_last_marker' has a default value of FALSE, meaning
 *   its value in the running process' settings is always ignored. Setting this
 *   argument implies items are added from an 'external' source and start() is
 *   not called to fetch them. start() may be called for preprocessing the
 *   passed items (depending on process settings), however the settings passed
 *   in this argument will not have effect during the call to
 *   _drunkins_run_start(); the running process' settings will be used there.
 *
 * @return bool
 *   TRUE if items were successfully fetched and queued (also if the number of
 *   items is 0); FALSE if the process state means this function should not be
 *   called or for error thrown during start() call.
 */
function _drunkins_enqueue_items($job_id, array $items = NULL, array $override_runner_settings = NULL) {
  $info = drunkins_job_status_info($job_id);
  if ($info['running'] != 3 || $info['method'] === 'batch') {
    _drunkins_log($job_id, "_drunkins_enqueue_items() called for job \"@job\" which is not running or is a 'batch' job.", array('@job' => $job_id), WATCHDOG_ERROR);
    return FALSE;
  }

  // 'skip_save_checkpoint' means do not save checkpoint statistics (start time
  // now, finish+counters when the context later) for the (de)activation of our
  // context. This isn't a real runner setting until now, mainly because we
  // haven't thought about whether it could be more widely useful.
  $context = &drunkins_current_context($job_id, !empty($override_runner_settings['skip_save_checkpoint']));
  unset($override_runner_settings['skip_save_checkpoint']);

  $runner_settings = _drunkins_get_process_config($job_id);
  if (isset($override_runner_settings)) {
    if (empty($info['can_add_items_while_running'])) {
      _drunkins_log($job_id, 'Job "@job" does not support adding extra items when already running.', array('@job' => $job_id), WATCHDOG_ERROR);
      return FALSE;
    }

    // Restrict the override settings to specific ones. In particular, we are
    // disallowing the use of 'enqueue_by_fetcher' - because 1) it's important
    // that we follow the value used in _drunkins_run_start(), and that function
    // cannot use our override value... and 2) changing it from the running
    // process' value has no known usefulness and is potentially confusing. Also
    // ignore 'add_last_marker' from the running process.
    unset($runner_settings['add_last_marker']);
    $filter = array_flip(array('logging', 'log_level', 'add_last_marker'));
    $runner_settings = array_intersect_key($override_runner_settings, $filter) + $runner_settings;
    $run_start = !empty($info['enqueue_by_fetcher']) || $info['can_add_items_while_running'] === 'pass_through_start' || $info['can_add_items_while_running'] === 'pass_through_start_by_default';
    if ($run_start) {
      // Pass $items through start() for preprocessing. (We assume no new items
      // will be fetched in start().)
      if (empty($info['can_override_fetch'])) {
        if (!empty($info['enqueue_by_fetcher'])) {
          _drunkins_log($job_id, "Job \"@job\" has no 'can_override_fetch' capability so it cannot handle the 'enqueue_by_fetcher' setting.", array('@job' => $job_id), WATCHDOG_ERROR);
        }
        else {
          _drunkins_log($job_id, "Job \"@job\" has inconsistent capabilities; items must be passed through start() but the job is not marked as being able to handle that (i.e. 'can_override_fetch' capability is not set).", array('@job' => $job_id), WATCHDOG_ERROR);
        }
        return FALSE;
      }
    }
    $can_loop = FALSE;

    // _drunkins_run_start() keeps the number of times start() was run, only if
    // specifically marked as 'start() needs rerun'. We will use the same
    // counter for logging purposes, regardless whether we actually run start()
    // or enqueue items directly.
    if (!isset($context['drunkins_rerun_start'])) {
      // Apparently start() was only run once and didn't need rerun. (Note that
      // when starting a new job from 'external' enqueued items, it also does
      // not get set; the 'else{}' block below gets executed instead.)
      // Set to 1 now.
      $context['drunkins_rerun_start'] = 1;
    }
    if (!$run_start) {
      // _drunkins_run_start() won't be executed so we'll do this ourselves:
      $context['drunkins_rerun_start']++;
    }
  }
  else {
    // 'first call' isn't always accurate; it's also TRUE when queueing extra
    // items received at finish().
    $run_start = !isset($items);
    if ($run_start && empty($context['drunkins_process_more'])) {
      // This process is not in a state where start() needs to be re-run, so
      // this function should not have been called.
      return FALSE;
    }

    // If this is a progressive job then we exit always; in that case if
    // $context['drunkins_process_more'] is nonempty, the next cron run / call
    // to the 'start again callback' should pick things up.
    // If this is the start of a non-progressive job, we might run again based
    // on $context['drunkins_process_more']. In practice this can only be the
    // case for a call where items were fed into this function; calling this
    // without $items for a process marked as 'non-progressive' is weird, given
    // the setup of the rest of the code.
    $can_loop = isset($runner_settings['progressive']) && empty($runner_settings['progressive']);
  }
  // If we run start() ourselves in the iteration below, we also want to reset
  // the memory limit (which _drunkins_run_start() can set, but does not reset
  // because the items are still in memory). If on the other hand the caller
  // has run _drunkins_run_start() by itself (regardless whether start() is
  // called again if the job is progressive), we leave it up to the caller to
  // reset it. And obviously if we don't run start() at all, we don't reset it.
  $reset_memory_limit = $run_start && (!empty($runner_settings['memory_limit_start']) || !empty($runner_settings['memory_limit']));

  do {
    if ($run_start) {
      if (isset($override_runner_settings)) {
        // Prepare for adding extra items.
        _drunkins_set_process_config($job_id, 'items', $items);
        if (!empty($runner_settings['enqueue_by_fetcher'])) {
          $count = count($items);
        }
      }

      try {
        $items = _drunkins_run_start($job_id, $context);
      }
      catch (\DrunkinsTemporaryError $e) {
        // Don't do anything with $e->getSuggestedWait() here; we assume our
        // caller doesn't immediately call us again. (It would be nice to
        // pass the exception with the suggested wait on to the caller but we
        // can only start thinking about that with a rewrite of the 'runner'.)
        _drunkins_log($job_id, 'Rerunning start for job "@job" threw temporary error; retrying: @error', array('@job' => $job_id, '@error' => $e->getMessage()), WATCHDOG_WARNING);
        return FALSE;
      }
      catch (\Exception $e) {
        // Process failed; is already canceled, queue cleaned up.
        _drunkins_log($job_id, 'Rerunning start for job "@job" threw exception; deleting all previously gathered items from the job. Exception: @error', array('@job' => $job_id, '@error' => $e->getMessage()), WATCHDOG_ERROR);
        return FALSE;
      }
    }

    if ($items && empty($runner_settings['enqueue_by_fetcher'])) {
      $add_last_marker = empty($context['drunkins_process_more']) && !empty($runner_settings['add_last_marker']);
      $queue = DrupalQueue::get($job_id);
      $nr = count($items);
      $i = 1;
      foreach ($items as $item) {
        // As the worker callback is the same for all Drunkins jobs, and
        // it can have only one argument, we need to include an ID for the
        // job (the queue name) _in_ the item.
        $qitem = array('data' => $item, 'queue_name' => $job_id);
        if ($add_last_marker && $i++ == $nr) {
          // This is just a hint to help finish the job as soon as possible, if
          // this item is processed last in the queue; there is no guarantee
          // that it is actually processed last. (There is a >99.9% chance of it
          // in most environments, except when exceptions are thrown during
          // processItem() calls.)
          $qitem['last'] = TRUE;
        }
        $queue->createItem($qitem);
      }
    }

    // Do logging; the message depends on a lot of factors (starting the
    // process; adding extra items externally; needing to add more; knowing
    // how many items we actually enqueued)

    if (empty($runner_settings['enqueue_by_fetcher'])) {
      $count = count($items);
      $count_param = $count;
      $total_param = $context['drunkins_rerun_start_count'];
    }
    else {
      // If isset($override_runner_settings) && !empty($runner_settings['enqueue_by_fetcher']),
      // $count was already set (because $items are destroyed now BUT we knew the
      // probable amount of items queued).
      if (!isset($override_runner_settings)) {
        $count = DrupalQueue::get($job_id)->numberOfItems();
      }
      // Part of the items may or may not already have been processed, and we
      // don't have the real value in context. Use non-text (i.e. no real
      // issues with translation) arg in messages to say "at least".
      $total_param = ">=$count";
      // Not a super nice solution inserting an extra space in the message but
      // let's worry about that later. Too many message permutations otherwise.
      $count_param = '';
    }

    if (!empty($context['drunkins_process_more'])) {
      if (isset($override_runner_settings)) {
        _drunkins_log($job_id, "While adding extra items to running job \"job_id\" and passing those though start(), the 'drunkins_process_more' context value was set. This is illegal and will be ignored.",
          array('@job' => $job_id), WATCHDOG_WARNING);
        unset($context['drunkins_process_more']);
      }
      elseif (empty($context['drunkins_continue_fetch_from'])) {
        _drunkins_log($job_id, 'Fetched @count items for job "@job" (total @total in @rerun_count iterations); trying to fetch more...',
          array('@rerun_count' => $context['drunkins_rerun_start'], '@count' => $count_param, '@total' => $total_param, '@job' => $job_id), WATCHDOG_INFO);
      }
      else {
        // drunkins_continue_fetch_from is used only here and can be set to any
        // desired value by a custom fetcher. It's not guaranteed to survive
        // into a next version; maybe there will be another way to influence
        // progress messages. For now, this hidden option will do.
        _drunkins_log($job_id, 'Fetched @count items for job "@job" (total @total in @rerun_count iterations); trying to fetch more, starting from @from...',
          array('@rerun_count' => $context['drunkins_rerun_start'], '@count' => $count_param, '@total' => $total_param, '@from' => $context['drunkins_continue_fetch_from'], '@job' => $job_id), WATCHDOG_INFO);
      }
    }
    else {
      if (isset($override_runner_settings)) {
        // 'drunkins_rerun_start' is the number of times start() has been run.
        // So the first time we enqueue extra items to a running process is
        // immediately marked '2' - or higher if start() was run multiple times.
        // This is the best / most consistent we can do.
        if (empty($runner_settings['enqueue_by_fetcher'])) {
          _drunkins_log($job_id, 'Added @count items to job "@job" from an external source (@rerun_count). ', array('@rerun_count' => $context['drunkins_rerun_start'], '@count' => $count, '@job' => $job_id), WATCHDOG_INFO);
        }
        else {
          _drunkins_log($job_id, 'Added (probably) @count items to job "@job" from an external source (enqueued by the job itself).', array('@count' => $count, '@job' => $job_id), WATCHDOG_INFO);
        }
      }
      elseif (!empty($context['drunkins_rerun_start'])) {
        // It's the last time this 'start again callback' is called.
        _drunkins_log($job_id, 'Fetched @count items for job "@job" (total @total in @rerun_count iterations); starting job.',
          array('@rerun_count' => $context['drunkins_rerun_start'], '@count' => $count_param, '@total' => $total_param, '@job' => $job_id), WATCHDOG_INFO);
      }
      elseif (_drunkins_get_process_config($job_id, 'items') !== FALSE) {
        // We're adding items for this process for the first time, so apparently
        // the items were added now. (The order of the outer if()s is such that
        // this would also still be true if above condition stayed true over
        // several HTTP requests, even though it doesn't at the moment.)
        if (empty($runner_settings['enqueue_by_fetcher'])) {
          _drunkins_log($job_id, 'Starting job "@job"; added @count items from an external source.', array('@count' => $count, '@job' => $job_id), WATCHDOG_INFO);
        }
        else {
          _drunkins_log($job_id, 'Starting job "@job"; added (probably) @count items from an external source (enqueued by the job itself).', array('@count' => $count, '@job' => $job_id), WATCHDOG_INFO);
        }
      }
      else {
        // This is the 'normal' situation where we are supposedly not called
        // from a 'start again callback' and $run_start was/will stay FALSE.
        if (empty($runner_settings['enqueue_by_fetcher'])) {
          _drunkins_log($job_id, 'Starting job "@job"; @count items to process', array('@count' => $count, '@job' => $job_id), WATCHDOG_INFO);
        }
        else {
          _drunkins_log($job_id, 'Starting job "@job"; @count items to process (enqueued by the job itself).', array('@count' => $count, '@job' => $job_id), WATCHDOG_INFO);
        }
      }
    }

    $run_start = !empty($context['drunkins_process_more']) && $can_loop;

  } while ($run_start);

  if ($reset_memory_limit) {
    _drunkins_set_memory_limit($job_id);
  }

  return TRUE;
}

/**
 * Update job info in database and status info, to say a process has finished.
 *
 * Optionally deletes the queue / batch(set) / items.
 *
 * @param string $job_id
 * @param array $options
 *   (Optional) options which influence behavior. (Even though the keys are
 *   long, the only use is writing self documenting internal code; we do not
 *   expect/want these options to be used by outside code.)
 *   'target_state' =>       string: can be 'manual_reset' / 'error_on_restart'
 *      / 'error_on_start' / 'empty_start' / 'error_on_finish'
 *      (If otherwise: target state will be 'normally finished'.
 *   'log_unknown_state'           boolean; default TRUE
 *   'delete_items'                boolean; default TRUE
 *   'job_may_restart_immediately' boolean
 *   (Note 'error_on_finish' really means 'stuck', not 'finished' and should
 *   always be accompanied by delete_items=FALSE. That is: this function is not
 *   only for "setting a process to finished state", but to "change a process
 *   from running to another state".)
 * @param array $job_context
 *   (Optional) Job context, if known. This is only used for optional actions
 *   (i.e. gathering statistics); if the context is lost, that's OK. The
 *   argument is primarily here because batch processes have to pass it.
 */
function _drunkins_finish_process($job_id, $options = array(), array $job_context = array()) {
  // Get job and context for later, while we still can.
  $job = drunkins_get_job($job_id);

  // Get updateable status info. (In some cases, like manual reset, this info
  // may not even be compiled yet and we end up wasting time. But at least this
  // way we are sure that the internally cached info is not stale.)
  $job_info = &drunkins_job_status_info($job_id);
  if (empty($job_context) && $job_info['method'] !== 'batch') {
    $job_context = drunkins_current_context($job_id);
  }

  // Compile the database fields to update.
  $fields = array();
  switch (isset($options['target_state']) ? $options['target_state'] : '') {
    // See drunkins_job_status_info() function comments for values.
    case 'manual_reset':
      // This is also the 'default' for true-ish non-string values.
      $fields['run_info'] = 5;
      break;
    case 'error_on_start':
    case 'error_on_restart':
      $fields['run_info'] = 4;
      break;
    case 'error_on_finish':
      if ($job_info['method'] !== 'queue') {
        // There is no reason that this could happen except a code error.
        _drunkins_log($job_id, 'Trying to set job "@job"\'s target stage "@target_state" but this does not exist for @method processes. The module code must be reviewed.', array('@job' => $job_id, '@target_state', 'error_on_finish', '@method' => $job_info['method']), WATCHDOG_CRITICAL);
        // Not setting $fields['run_info'] means we don't update state but do
        // delete items.
        $options['delete_items'] = TRUE;
      }
      else {
        $fields['run_info'] = 3;
      }
      break;
    case 'empty_start':
    default:
      // Mark as normally finished.
      $fields['run_info'] = 0;
  }
  // Check if the info says that the process is actually running (or if we are
  // resetting a queue job it may also be in a 'stuck' state).
  // Otherwise, log the inconsistency and exit.
  if (!isset($job_info['run_info'])
      || !($job_info['run_info'] == 0 && $job_info['started'] > $job_info['finished']
           || $job_info['run_info'] == 1 || $job_info['run_info'] == 2
           || $job_info['run_info'] == 3 && isset($options['target_state']) && $options['target_state'] === 'manual_reset'
  )) {
    // We don't know what to do on failure, or what kind of sensible message to
    // give to the end user. This isn't expected to have a huge effect, though
    // it's a sign of stale code. So log only to watchdog, a warning that is
    // only readable by the programmer.
    // (This function is not supposed to be called by outside code.)
    if (!isset($options['log_unknown_state']) || !empty($options['log_unknown_state'])) {
      watchdog('drunkins', "Unknown process 'run_info' state \"@strange_state\" while calling _drunkins_finish_process() to set run_info=@new_info. The module code should be reviewed.",
        array('@strange_state' => $job_info['run_info'], '@new_info' => $fields['run_info']), WATCHDOG_WARNING);
    }
  }
  elseif (isset($fields['run_info'])) {
    // Compile the database fields to update.
    if (isset($options['target_state']) && $options['target_state'] === 'manual_reset') {
      // We don't want to leave a big blob of context data behind in an unknown
      // state. We may keep it in case the job was normally finished (or in case
      // of error, this could be useful for inspection?) but in this case
      // there's a higher chance of 'dangerous' data leaking.
      $info = db_query("SELECT data FROM {drunkins_process} WHERE queue_name = :qn",array(':qn' => $job_id))
        ->fetchAssoc();
      $data = unserialize($info['data']);
      unset($data['context']);
      $fields['data'] = serialize($data);
    }

    // 'error_on_finish' is not really 'finished' so don't change the time.
    if (!isset($options['target_state']) || $options['target_state'] !== 'error_on_finish') {
      $fields['finished'] = time();
      // If a new process could start immediately after this function exits, we
      // could have an issue when 'started' gets set to the current time
      // because 'started == finished' means a job is not running. So make sure
      // that 'finished' will be smaller than that theoretical start time by
      // decrementing it.
      // @todo while OK in practice, this still seems flakey. Maybe change our assumption elsewhere in the code, that only 'started > finished' means a job is running?
      if (!empty($options['job_may_restart_immediately'])) {
        if ($job_info['started'] == $fields['finished']) {
          // Theoretical edge case: by decreasing 'finished', the start time
          // would become greater so the process would seem to be running. Fix.
          $fields['started'] = $fields['finished'] - 1;
        }
        $fields['finished']--;
      }
    }

    db_update('drunkins_process')
      ->fields($fields)
      ->condition('queue_name', $job_id)
      ->execute();

    // Make sure the cached 'job info' is up to date too, to prevent obscure
    // bugs. (At least drunkins_cron_info() needs up to date 'running' info.)
    $job_info = array_merge($job_info, $fields);
    _drunkins_calculate_dynamic_info($job_info, $job_id);
  }

  // Save process counts. It's unfortunate we need to get a process instance
  // for this, though it should be fine. (We'll move saveProcessSummaryCounts()
  // elsewhere later, when possible.)
  if ($job_context && variable_get('drunkins_save_process_summary', FALSE)) {
    if (empty($job_context['drunkins_process_id'])) {
      // This is not fatal for our purpose but it's certainly strange.
      watchdog('drunkins', "'drunkins_process_id' not found in context for job \"@job\"; cannot save statistics.", ['@job' => $job_id], WATCHDOG_WARNING);
    }
    else {
      db_update('drunkins_process_run')->fields(['finished' => time()])
        ->condition('id', $job_context['drunkins_process_id'])->execute();
      if (is_callable(array($job, 'saveProcessSummaryCounts'))) {
        $job->saveProcessSummaryCounts($job_context['drunkins_process_id'], FALSE, $job_context);
      }
    }
  }

  // Remove the class instance so it will be reinstantiated if a new job
  // with the same ID gets started. In principle this is unnecessary
  // because the instance should not hold any state information, so we
  // should be able to reuse it... But we (arbitrarily) draw a line at
  // reusing class instances over different processes, to prevent hard to
  // trace bugs from class implementations that do not follow the rules.
  // (Bugs occurring during the lifetime of one process are supposedly less
  // hard to spot.)
  unset($job_info['process_instance']);

  // Delete items at the end; this is the expensive/least predictable operation.
  // It never makes sense to not delete items. However since this is an internal
  // function, and we in theory don't know how expensive the queue mechanism is,
  // a call can choose to not delete them if it knows they're not there.
  // (At this point, the drunkins cron-queue code would discard any queued
  // items it received without context. But purging the queue is neater.)
  if (!isset($options['delete_items']) || !empty($options['delete_items'])) {
    DrupalQueue::get($job_id)->deleteQueue();
  }
}

/**
 * Runs a queue process in a 'non-progressive' way, i.e. processes all items
 * in this one function call. Please see comments at drunkins_start_job() for
 * caveats; the concept of Drunkins jobs is not designed for doing this.
 *
 * There is no return value; drunkins_last_log() will contain the message
 * / context from finish(). (We assume that drunkins_last_log() at this moment
 * returns a value which does not contain context, so that a succesfully-fully
 * run process can be distinguished by that.)
 *
 * There's no reason for this to be a separate function except
 * drunkins_start_job() gets too long.
 *
 * @param $job_id
 *   Job ID / machine name for the the queue.
 * @param string $method
 *   Method used for running this job.
 *
 * @see drunkins_start_job()
 */
function _drunkins_process_queued_items($job_id, $method) {

  // Initialize last-log; we'll check that to see if the process finished.
  // It's near-impossible that it has 4 values now, but:
  $last_log = &drupal_static('drunkins_last_log');
  unset($last_log[3]);

  // Get pointer into 'live' context. We don't want to change anything in it.
  $job_context = &drunkins_current_context($job_id);

  $runner_settings = _drunkins_get_process_config($job_id);
  $progress_interval = !empty($runner_settings['update_progress_interval']) ? $runner_settings['update_progress_interval'] : 50;

  $queue = DrupalQueue::get($job_id);
  while (TRUE) {

    // Simulate a drupal_cron_run() that doesn't check execution time.
    while ($item = $queue->claimItem()) {
      _drunkins_cron_process_queue_item($item->data);
      $queue->deleteItem($item);

      if ($progress_interval && isset($job_context['drunkins_progress']) && is_numeric($job_context['drunkins_progress'])
          && ($job_context['drunkins_progress'] % $progress_interval) == 0 && $method === 'drush' && !drush_get_context('DRUSH_QUIET')) {
        // Show update progress. This is not a normal log message; it should
        // only be shown on the display for some methods - drush for now. We
        // will consider a drush progressbar library when we start using
        // composer.
        drush_print(t("processed @nr items\r", array('@nr' => $job_context['drunkins_progress'])), 0, STDERR, FALSE);
      }
    }

    // Doublecheck whether claimItem() failed on us.
    if ($queue->numberOfItems()) {
      _drunkins_log($job_id, 'While running job "@job": claimItem() returned false with items still queued. Aborting further processing.', array('@job' => $job_id), WATCHDOG_ERROR);
      // In the odd case that $return is still filled (which means the job
      // finished properly even though now there are suddenly items in the queue
      // which is only possible in some strange race condition), just return it.
      break;
    }

    if (isset($last_log[3])) {
      // 'last' item triggered a call to finish(); normal exit.
      break;
    }

    // 'last' item did not trigger a call to finish(); do it now.
    _drunkins_queue_process_check_finish($job_id);
    // This may cause the queue to be re-filled. If so, jump back in the loop.
    if ($queue->numberOfItems()) {
      continue;
    }

    if (!isset($last_log[3])) {
      $log_source = array();
      if ($runner_settings['logging'] & DrunkinsJob::LOG_SCREEN) {
        $log_source[] = 'on the screen';
      }
      if ($runner_settings['logging'] & DrunkinsJob::LOG_WATCHDOG) {
        $log_source[] = 'in watchdog logs';
      }
      _drunkins_log($job_id, "While running job \"@job\": finish() did not run correctly. See earlier messages !source for details.",
        array('@job' => $job_id, '!source' => implode('/', $log_source)), WATCHDOG_ERROR);
    }
    break;
  }

}

/**
 * Worker callback (1 item) for cron queue process, as called from .module.
 */
function _drunkins_cron_process_queue_item(array $qitem) {

  if (isset($qitem['queue_name']) && $qitem['queue_name'] !== '') {

    // If we end up here (i.e. there are queued items) but the job status is
    // 'not running' or context is empty, then discard items; we can't run
    // reliably without this info. This means that deleting a row in
    // {drunkins_process} is equivalent to canceling the running job.

    $context = &drunkins_current_context($qitem['queue_name']);
    if ($context) {
      $job = drunkins_get_job($qitem['queue_name']);
      if ($job) {
        // Process one item.
        // 'progress' gets incremented every time before an item is processed.
        // (This means it's not accurate in case of parallel processing, but
        // then again, this whole module can't deal with parallel processing
        // because $context can lose info too.)
        $context['drunkins_progress']++;
        try {
          $job->processItem($qitem['data'], $context);
        }
        catch (\Exception $e) {
          $runner_settings = _drunkins_get_process_config($qitem['queue_name']);
          $handle = $e instanceof \DrunkinsTemporaryError && isset($runner_settings['handle_process_temp_error'])
            ? $runner_settings['handle_process_temp_error']
            : (isset($runner_settings['handle_process_exception']) ? $runner_settings['handle_process_exception']
              : 'fallthrough');
          if ($handle !== 'fallthrough') {
            // We have various (12?) permutations for $log_message, which is
            // icky for translation but there isn't a much nicer way. There's
            // no dot after "@message" because we assume the error message
            // contains one.
            $log_message = $e instanceof \DrunkinsTemporaryError ? 'encountered a temporary error' : 'failed';
            if ($handle === 'discard') {
              $log_message .= '; discarding item';
            }
            $log_message = "Processing '@job' item $log_message: @message";
            $log_params = ['@job' => $qitem['queue_name'], '@message' => $e->getMessage()];
            $log_item = $e instanceof \DrunkinsTemporaryError && isset($runner_settings['process_temp_error_log_item'])
              ? $runner_settings['process_temp_error_log_item']
              : (isset($runner_settings['process_exception_log_item']) ? $runner_settings['process_exception_log_item']
                : FALSE);
            if ($log_item) {
              // Our log (that is, watchdog or drupal_set_message) can contain
              // HTML for formatting and a LF doesn't work as a newline. Also,
              // in order to have our _substituted values_ come out with
              // newlines 'working', we'll modify the _message_ to include
              // <pre>. That isn't future proof / nice to other logging
              // backends, but that's the way it is for now.
              // @todo when we make a logger, do an option to convert LF to <br>
              //   in the above backends that don't handle LF? And make some
              //   replacement for the <pre> mechanism?
              $json_pretty_format = stripos($log_item, 'json') === 0 ? (bool) stripos($log_item, 'pretty') : NULL;
              $log_message .= $json_pretty_format === FALSE ? '<br>Item: @item' : '<br>Item: <pre>@item</pre>';
              $log_params['@item'] = $json_pretty_format === NULL
                ? var_export($qitem['data'], TRUE) : json_encode($qitem['data'], $json_pretty_format ? JSON_PRETTY_PRINT : 0);
            }
            if ($e instanceof \DrunkinsTemporaryError && isset($runner_settings['process_temp_error_log_trace'])
              ? $runner_settings['process_temp_error_log_trace']
              : (isset($runner_settings['process_exception_log_trace']) ? $runner_settings['process_exception_log_trace']
                : FALSE)) {
              $log_message .= '<br>Trace: <pre>@trace</pre>';
              $log_params['@trace'] = $e->getTraceAsString();
            }
          }
          if ($handle === 'discard') {
            _drunkins_log($qitem['queue_name'], $log_message, $log_params, WATCHDOG_ERROR);
            // Do accounting, and wait if the exception indicates it. The item
            // will be dequeued.
            if ($e instanceof \DrunkinsTemporaryError) {
              $context['results']['drunkins_temp_error_count']++;
              // It would be best if we stopped processing this queue at all
              // and continue with something else useful. We can't influence
              // that though; we have to assume our caller will keep calling us
              // with new items to process, for whatever time period we
              // specified in hook_queue_info(). So let's sleep() for at least
              // that long, if the suggested wait time is longer.
              // @todo this can wait for <queue time> also when we've already
              //   been processing for nearly <queue time> and then get an
              //   error - meaning we could be in this state for almost twice
              //   the reserved queue time. Do something about that by
              //   recording/incorporating the 'start time' in whatever way.
              $sleep = min($e->getSuggestedWait(), _drunkins_queue_time($qitem['queue_name']));
              if ($sleep) {
                if ($sleep) {
                  _drunkins_log($qitem['queue_name'], 'Waiting @wait seconds...', array('@wait' => $sleep), WATCHDOG_INFO);
                }
                sleep($sleep);
              }
            }
            else {
              // A regular exception has no 'suggested wait time' here and we
              // don't by default assume that there is anything time based
              // about whatever caused the exception to be thrown. So continue
              // immediately.
              $context['results']['drunkins_exception_count']++;
            }
          }
          else {
            // 'log' / 'fallthrough'. (Other values should have been caught
            // before process start.)
            if ($handle === 'log') {
              // We have to log more than just the message.
              _drunkins_log($qitem['queue_name'], $log_message, $log_params, WATCHDOG_ERROR);
            }

            // UNDOCUMENTED functionality until we have a PSR-3 logger in place.
            $job_settings = _drunkins_get_process_config($qitem['queue_name'], 'job');
            if (isset($job_settings['log_extra_function']) && is_string($job_settings['log_extra_function'])
                && function_exists($job_settings['log_extra_function'])) {
              $job_settings['log_extra_function']($e->getMessage(), [], WATCHDOG_ERROR, $job_settings);
            }

            throw $e;
          }
        }
      }
    }

    if (!empty($qitem['last'])) {
      if ($context) {
        // 'last' means it was the last item queued for this job originally,
        // which means it's _probably_ the last item processed. This is the only
        // case where we execute DrupalQueue::get($queue_name)->numberOfItems()
        // to check if we can finish the job within this worker callback; it is
        // considered to be too expensive to call inside every callback.
        // If the job is not finished now, it will get finished at the beginning
        // of the next cron run after all items have already been processed.
        // (This means that a system which calls its own worker callbacks, will
        // also have to make sure that the job gets finished, after all items
        // are processed.)
        // The last item will automatically be removed after returning from this
        // worker callback, so a job's finish() will be called _before_ the
        // queue is empty. This should not matter (finish() has no knowledge of
        // the queue mechanism).
        _drunkins_queue_process_check_finish($qitem['queue_name'], array('last_item_still_in_queue' => TRUE));
      }
      else {
        // $context was deleted at some point and probably all queued items fell
        // through silently. Log a warning for this last one.
        _drunkins_log($qitem['queue_name'], 'Completed queued job "@job": @message', array(
            '@job' => $qitem['queue_name'],
            '@message' => t('Several items were discarded because context / run status was invalid.')
          ), WATCHDOG_WARNING);
      }
    }
  }
}

/**
 * Returns the processing time for a given queue.
 */
function _drunkins_queue_time($queue_name) {
  global $_elysia_cron_settings;
  if (isset($_elysia_cron_settings) && isset($_elysia_cron_settings["queue_$queue_name"]['arguments'][1]['time'])) {
    // 'live' time, may have been modified by drunkins_cron().
    $time = $_elysia_cron_settings["queue_$queue_name"]['arguments'][1]['time'];
  }
  else {
    // Get it from cron-queue info.
    $queues = drunkins_cron_queue_info();
    $time = isset($queues[$queue_name]['time']) ? $queues[$queue_name]['time'] : 15;
  }

  return $time;
}

/**
 * Check if a queue-mode process* which is marked 'running' has processed all
 * items, and if it has, finish it. Note a possible effect of this is that the
 * queue is repopulated with extra items, returned by finish().
 *
 * *meaning: a process which uses the Drupal Queue mechanism to process its
 * items. This can be more than just the 'queue' method.
 *
 * Supported options which influence behavior: (Even though the keys are long,
 * the only use is writing self documenting internal code; we do not
 * expect/want these options to be used by outside code.)
 *  'last_item_still_in_queue' (boolean)
 *  ...any options supported by _drunkins_finish_process(), a.o.:
 *    'job_may_restart_immediately' (boolean)
 */
function _drunkins_queue_process_check_finish($job_id, $options = array()) {

  // If job isn't running, silently skip it.
  $info = &drunkins_job_status_info($job_id);
  if ($info['running'] != 3 || $info['method'] === 'batch') {
    // We may decide later to delete this if(); it really should never happen.
    _drunkins_log($job_id, 'Current method @method or run_info code @run_info of job "@job" is illegal when calling _drunkins_queue_process_check_finish(). This indicates a code error.', array('@method' => $info['method'], '@run_info' => $info['run_info'], '@job' => $job_id), WATCHDOG_CRITICAL);
  }
  // If 'drunkins_process_more' is set, skip it. If context is empty (which
  // should never happen) then 'finish' it too but log an error.
  elseif (empty($info['process_context']['drunkins_process_more'])) {
    // Recheck if the number of items is really 0 (i.e. the job is really
    // finished) and exit silently if it isn't. (Some callers use this feature.)
    // A potential race condition exists because an item may be added just
    // after checking that the queue is empty. This item will be discarded. If
    // you're worried about this, then use DrunkinsOrderedQueue.
    // @todo implement DrunkinsOrderedQueue which does not accept new items once empty - also should process items in exact order.
    $items = DrupalQueue::get($job_id)->numberOfItems();
    if (!empty($options['last_item_still_in_queue']) ? $items < 2 : !$items) {

      $runner_settings = array();
      $context = &drunkins_current_context($job_id);
      if ($context) {
        $job = drunkins_get_job($job_id);
        if ($job) {
          // Log before finish() (at debug level).
          // @todo what do we exactly want to do immediately after, when finish() does not return a message anymore?
          _drunkins_log($job_id, 'Calling finish() for job "@job".', array('@job' => $job_id), WATCHDOG_DEBUG);
          $runner_settings = _drunkins_get_process_config($job_id);
          if (!empty($runner_settings['memory_limit_finish'])) {
            _drunkins_set_memory_limit($job_id, $runner_settings['memory_limit_finish']);
          }
          //@todo think: is there a chance of a race condition when another process is
          //   adding an item to the queue of this running job, while we are already
          //   running finish()? What to do about it?
          //@todo think: is there a chance of a race condition when we run
          //   _drunkins_finish_process() but the last queue item still needs to
          //   be dequeued (by the cron code)? If so, what to do about it?
          //   If not, maybe make a comment somewhere.
          try {
            $message = $job->finish($context);

            if (!empty($context['drunkins_process_more']) && is_array($context['drunkins_process_more'])) {
              // If we got any message back, don't just drop it; log/print it.
              if ($message) {
                _drunkins_log($job_id, $message, array(), WATCHDOG_INFO);
              }
            }
            else {
              // Log summary message about processItem() throwing exceptions.
              // (Each has logged an individual error but it seems good to do a
              // summary - with WARNING severity then.)
              if (!empty($job_context['drunkins_exception_count'])) {
                _drunkins_log($job_id, '@count exceptions were thrown while processing items for job "@job".',
                  array('@job' => $job_id, '@count' => $job_context['drunkins_exception_count']), WATCHDOG_WARNING);
              }
              _drunkins_log($job_id, 'Completed queued job "@job": @message', array('@job' => $job_id, '@message' => $message), WATCHDOG_INFO);
              // TODO display message/context/settings in drunkins UI/overview.

              // Last log was stored, but also store context as a signal that the
              // process was successfully finished.
              $log = &drupal_static('drunkins_last_log');
              $log = array($message, array(), WATCHDOG_INFO, $context);
            }
          }
          catch (Exception $e) {
            //@TODO if we make the items the return value, we can simplify a lot below because we cannot generate an exception AND more items.
            //  We'll also have to think about the other @todo below though.
            if (isset($runner_settings['handle_finish_exception'])
                && ($runner_settings['handle_finish_exception'] === 'continue'
                    || $runner_settings['handle_finish_exception'] === 'finish')) {
              // Log exception and finish normally (or continue on more items).
              _drunkins_log($job_id, 'Calling finish() for job "@job" failed: @message', array('@job' => $job_id, '@message' => $e->getMessage()), WATCHDOG_ERROR);
              if (!empty($job_context['drunkins_process_more']) && is_array($job_context['drunkins_process_more'])) {
                if ($info['handle_finish_exception']['continue_on_finish_exception'] === 'continue') {
                  _drunkins_log($job_context, '"@job" finish() generated @count more items to process: continuing despite exception thrown.', array('@job' => $job_id), WATCHDOG_NOTICE);
                }
                else {
                  _drunkins_log($job_id, '"@job" finish() generated @count more items to process: discarding them because of exception thrown.', array('@job' => $job_id), WATCHDOG_NOTICE);
                  unset($job_context['drunkins_process_more']);
                }
              }
            }
            else {
              // Default value: 'break'. (Other values should have been caught
              // before process start.)

              // This function is called in two situations:
              // - from a cron queue worker: we cannot rethrow the exception
              //   because the last item is still queued and we must let the
              //   caller dequeue it (so it will not get re-processed). We have
              //   to just log a message. (And then we could potentially just
              //   keep the process state "running" which would get us to the
              //   next situation some time later).
              // - from some other place that checks if the job is registered
              //   as status 'running' but has 0 items. At least one caller can
              //   potentially do this quite often: drunkins_cron_queue_info().
              //   Since we don't want to introduce a situation where a failing
              //   finish() is called repeatedly and indefinitely in the
              //   background... we need to introduce an extra 'error on finish'
              //   state (which either us or the caller should set).
              // For uniformity's sake, we won't throw and will move the process
              // to this state in both above cases. Note this 'error on finish'
              // state will only exist for queue processes because for batch,
              // there is no risk of constantly rerunning finish / the runner
              // will just throw an exception which the caller handles in a way
              // we like, if we keep the process status the same. (See comments
              // at drunkins_batch_finished().) Note also that there is no way
              // to rerun finish() for a process in 'error on finish' state
              // (unlike a standard interactive batch process where you can just
              // reload the page); a process can only be completely reset.
              // @todo make rerunning finish() possible, if we ever care.
              _drunkins_log($job_id, 'Calling finish() for job "@job" failed: @message', array('@job' => $job_id, '@message' => $e->getMessage()), WATCHDOG_ERROR);
              // We're not logging about any extra items; they *probably* will
              // be saved into context (if the PHP 'shutdown functions' still
              // run after an exception; haven't tested that). Potentially they
              // could be processed if the user reloads and triggers finish()
              // and that finishes without throwing an exception next time.
              // There is no facility for that at the moment, though.
              // @todo this completely changes if we make the extra items a return
              //   value instead of a subkey in $context. Think about this.

              $options['target_state'] = 'error_on_finish';
              $options['delete_items'] = FALSE;
            }
          }
        }
        else {
          // This should never happen.
          _drunkins_log($job_id, 'Failed to create class instance for job "@job" while trying to finish it!?', array('@job' => $job_id), WATCHDOG_ERROR);
          $options['target_state'] = 'error_on_finish';
          $options['delete_items'] = FALSE;
        }
      }
      else {
        _drunkins_log($job_id,
          '_drunkins_queue_process_check_finish() called without context for job "@job"; context was apparently missing from the database table?',
          array('@job' => $job_id, '@c' => $info['runner_settings']['class']),
          WATCHDOG_ERROR);
      }

      // @todo consider if we want to make this array (part of) the return
      // argument, instead of the message (if we want to get rid of that message as return argument anyway).
      if (!empty($context['drunkins_process_more']) && is_array($context['drunkins_process_more'])) {
        // @todo maybe make the 'executive decision' that status logs like the below
        // will not be done by the runner, and logging is the exclusive responsibility
        // of the job (except in clear errors)? Not sure; we should treat it the same as the 'N items received' from start().
        _drunkins_log($job_id,
          'finish() for job "@job" returned @count new items; continuing...',
          array('@job' => $job_id, '@count' => count($context['drunkins_process_more'])),
          WATCHDOG_INFO);
        // Enqueue items and continue. This also resets memory limit. Empty
        // $context['drunkins_process_more'], since that might confuse the
        // queueing code to call start() again internally.
        $items = $context['drunkins_process_more'];
        unset($context['drunkins_process_more']);
        _drunkins_enqueue_items($job_id, $items);
        // Unset 'drunkins_process_more' so hook_cron(_info)() won't assume
        // start() needs to be called again, and to free memory.
        unset($context['drunkins_process_more']);
        // Reset memory limit from the 'finish' one, if applicable. If there is
        // a general memory_limit setting, we will set that rather than the
        // initial one, since our caller may continue processing items.
        if (!empty($runner_settings['memory_limit_finish'])) {
          $initial_limit = _drunkins_get_initial_memory_limit();
          _drunkins_set_memory_limit($job_id, $initial_limit > $runner_settings['memory_limit'] ? $initial_limit : $runner_settings['memory_limit'], TRUE);
        }
      }
      else {
        _drunkins_finish_process($job_id, $options + array('delete_items' => FALSE));
        // We have a slight preference for emptying out the static info that
        // isn't needed anymore, but then we have to explicitly deactivate the
        // context first - which would be done from a shutdown function - so it
        // can do e.g. any saving of statistics.
        drunkins_current_context('');
        $context = array();
      }
    }
  }
}

/**
 * Returns a class instance for a job.
 *
 * Calls with and without the $settings argument serve a somewhat different
 * purpose. For $settings: see below. When not passing $settings: if the job is
 * not running, the returned instance is initialized with the standard settings
 * as defined in hook_queue_info(). If it is running, the instance is
 * initialized with the settings from the running process.
 * (That is: the settings which were passed to e.g. drunkins_start_*_job() to
 * start the process, plus the standard settings.)
 *
 * This function does nothing about starting jobs; it only gets instances for
 * the current job state. It shouldn't be as 'user facing' as it looks now. (In
 * fact it's only called from .run.inc and might be moved there.)
 *
 * @param string $job_id
 *   Job ID, also used as machine name of the queue for a cron/queue process
 * @param array $settings
 *   Settings influencing this specific job. When passing this argument, running
 *   jobs are ignored and the returned job instance will be initialized with
 *   $settings plus the standard settings from hook_queue_info(). This argument
 *   is passed only by code which is sure that a job is not running, or wants to
 *   use a job class outside of the regular 'process flow'.
 *
 * @return NULL|DrunkinsJob
 *   Class instance ready for use, or NULL if something went wrong.
 *
 * @todo investigate in how far we can get rid of the $settings argument, and if
 *   we can unify their behavior more with how 'runner settings' are currently done:
 *   store them in _drunkins_start_process() and never call this function before
 *   that time. Not sure how well that sits with the fringe cases where we need
 *   this function.
 */
function drunkins_get_job($job_id, array $settings = NULL) {

  $job_info = &drunkins_job_status_info($job_id);
  if ($job_info) {
    if (isset($job_info['process_instance']) && !isset($settings)) {
      return $job_info['process_instance'];
    }

    if ($job_info['running'] == 3 && !isset($settings)) {
      // Get settings from the running process.
      if (!isset($job_info['process_runner_settings']) || !is_array($job_info['process_runner_settings'])) {
        return NULL;
      }
      $runner_settings = $job_info['process_runner_settings'];

      if (!isset($job_info['process_settings']) || !is_array($job_info['process_settings'])) {
        return NULL;
      }
      $job_settings = $job_info['process_settings'];
    }
    else {
      // Don't get settings from the running process.

      if (!isset($job_info['runner_settings']) || !is_array($job_info['runner_settings'])) {
        // If settings is not an array it is defined wrongly, or modified (by
        // reference) by buggy code. We won't run with buggy settings and need
        // at least 'class' below.
        // @todo log?
        return NULL;
      }
      $runner_settings = $job_info['runner_settings'];

      if (!isset($job_info['settings'])) {
        $job_settings = array();
      }
      elseif (!is_array($job_info['settings'])) {
        // @todo log?
        return NULL;
      }
      else {
        $job_settings = $job_info['settings'];
      }
      if (isset($settings)) {
        $job_settings = array_merge($job_settings, $settings);
      }
    }

    if (!isset($runner_settings['class'])) {
      // @todo once we've implemented logger, see how to replace watchdog calls.
      watchdog('drunkins', "drunkins_job_status_info() has no 'class' definition for job \"@job\"",
        array('@job' => $job_id), WATCHDOG_ERROR);
    }
    elseif (!is_string($runner_settings['class'])) {
      watchdog('drunkins', "drunkins_job_status_info() has invalid 'class' definition for job \"@job\": @c",
        array('@job' => $job_id, '@c' => print_r($runner_settings['class'], TRUE)), WATCHDOG_ERROR);
    }
    else {
      // Add any 'runner settings' that must be propagated to the job
      // class because they are also supported there. (This is kind-of hackish
      // and dependent on hardcoded logic in default jobs.)
      // - 'fetcher_class' is copied into settings as 'drunkins_fetcher_class'.
      // - 'enqueue_by_fetcher' is copied into settings as
      //   'drunkins_enqueue_by_fetcher'; this is likely a temporary measure.
      // @todo to formally support this, we will need to inject something (maybe a Queue, maybe a Runner class)
      //   through context. At that time we might decide to move 'enqueue_by_fetcher' to context too...
      // - 'logging' is copied into settings as 'logging'.
      // Define an extra variable because we don't want to change the
      // 'process_runner_settings' which are stored in status info.
      $settings_not_stored_in_info = array();
      if (isset($runner_settings['fetcher_class']) && is_string($runner_settings['fetcher_class'])) {
        $settings_not_stored_in_info['drunkins_fetcher_class'] = $runner_settings['fetcher_class'];
      }
      if (!empty($runner_settings['enqueue_by_fetcher'])) {
        $settings_not_stored_in_info['enqueue_by_fetcher'] = $runner_settings['enqueue_by_fetcher'];
      }
      if (isset($runner_settings['logging'])) {
        $settings_not_stored_in_info['logging'] = $runner_settings['logging'];
      }
      // Also add job_id.
      $settings_not_stored_in_info['job_id'] = $job_id;
      try {
        $instance = new $runner_settings['class']($job_settings + $settings_not_stored_in_info, $job_id);
        // Cache the job instance if it's created for a running process.
        if (!isset($settings) && $job_info['running'] == 3) {
          $job_info['process_instance'] = $instance;
        }
        return $instance;
      }
      catch (Exception $e) {
        watchdog('drunkins', 'drunkins_get_job() failed to instantiate class @c for job "@job": @e',
          array('@job' => $job_id, '@c' => $runner_settings['class'], '@e' => $e->getMessage()), WATCHDOG_ERROR);
      }
    }
  }
}

/**
 * Returns a reference to a running non-batch process' active context.
 *
 * This is the (mandatory) way to obtain a reference to context data by code
 * that might alter the process context. This function takes care of 'lazily'
 * saving the context to persistent storage whenever a new process is switched
 * to or when the PHP process exits, and remembers current/up to date context
 * in the meantime.
 *
 * This is not supported for 'batch' method because that has its own way of
 * handling context.
 *
 * @param string $queue_name
 *   The ID of the job whose context must be returned; after the first call with
 *   this particular job id, the function will be responsible for saving its
 *   context at a later point. The job must be running in queue mode; otherwise
 *   an empty array is returned. '' explicitly lets the active context be saved
 *   and become 'inactive', and returns an empty array.
 * @param bool $suppress_new_checkpoint
 *   (Optional) do not create a new 'checkpoint' row in the database for the
 *   new job, which would store the start time and be updated with finished +
 *   all counters when the new context goes out of scope. The currently
 *   in-scope context will still have its data saved irrespective of this
 *   parameter - if 'checkpoint saving' is enabled. Otherwise this parameter
 *   has no effect because nothing is saved regardless of its value.
 *
 * @return array
 *   The context data.
 */
function &drunkins_current_context($queue_name, $suppress_new_checkpoint = FALSE) {
  // @todo consider storing only the settings overrides in the database;
  //   not the queue info. Better for security/privacy, e.g. no passwords stored
  //   but on the other hand that could mean that if a job config gets changed,
  //   its settings/runner settings change too. Think about this. Config option?

  static $current_queue;

  // Make sure not to change anything / register the shutdown function if
  // $current_queue is still NULL and $queue_name is ''. The below will also
  // not do anything if the queue name is 0/"0" and... we're fine with that.
  $change = $queue_name != $current_queue;

  if (!isset($current_queue) && $queue_name) {
    // This is the first queue item from the first job in this HTTP request.
    // We don't need to write an existing cache yet, but make sure this is
    // done on shutdown.
    drupal_register_shutdown_function('drunkins_flush_info');
  }

  if ($current_queue && $change) {
    // Write the current context to a persistent cache, before starting to
    // work on a new queue.
    // We use our dedicated table for this:
    // - no variable_set(); this clears caches;
    // - no ctools_object_cache_set because that depends on session info.
    // Note we might also wait until the PHP process ends and save all changed
    // processes' info to persistent cache at shutdown. However, writing
    // whenever another process 'becomes current' seems like a reasonable
    // tradeoff between speed and safety of not losing data.

    $info = drunkins_job_status_info($current_queue);

    // If context is emptied somehow, we assume the job doesn't exist anymore
    // and the database info has already been updated.
    // @todo this shouldn't happen though. Should we log once per PHP exec?
    if (!empty($info['process_context'])) {
      // As a counterpart to _drunkins_queue_info_alter(), collect only known
      // properties that should be stored in the 'data' database field:
      $data = array(
        'context' => $info['process_context'],
        'settings' => $info['process_settings'],
      );
      if (!empty($info['process_runner_settings'])) {
        $data['runner_settings'] = $info['process_runner_settings'];
      }
      // Do a simple update query; if another process just deleted our info,
      // the context will not be saved and processing will stop.
      $updated = db_update('drunkins_process')
        ->fields(array('data' => serialize($data)))
        ->condition('queue_name', $current_queue)
        ->execute();
      if (!$updated) {
        // One possible cause of the update failing is the serialized data
        // being too big. Set empty context; if a database row indeed still
        // exists, this will cause all items to be discarded, which could be
        // better than processing them with stale context.
        // This fact will be logged in _drunkins_cron_process_queue_item().
        db_update('drunkins_process')
          ->fields(array('data' => serialize(array())))
          ->condition('queue_name', $current_queue)
          ->execute();
        // Below is a possibly-temporary measure using a variable that is not
        // exposed anywhere. If set, e-mail someone about the failure because
        // this is potentially serious; if context data is lost quietly then
        // (depending on the use case) processes running in the background
        // could stop running quietly. For being warned about this with a
        // (very basic text) e-mail, set an e-mail address here. This pairs
        // with the check in drunkins_flush_info().
        if ($mailto = variable_get('drunkins_fatal_error_mailto', '')) {
          // Setting '..._include_context' variable is security sensitive.
          _drunkins_mail_fatal_error($mailto, "Failed to save context for $queue_name; the process will likely abort."
            . (variable_get('drunkins_fatal_error_mail_include_context', FALSE)
              ? serialize($data) : ''));
        }
      }

      // Check if we should be saving process 'checkpoint' statistics on each
      // switch.
      if (!empty($info['process_context']['drunkins_process_sub_id']) && variable_get('drunkins_process_summary_checkpoint', FALSE)) {
        db_update('drunkins_process_run')->fields(['finished' => time()])
          ->condition('id', $info['process_context']['drunkins_process_sub_id'])->execute();
        // Too bad we need an instance for this at the moment...
        $job = drunkins_get_job($current_queue);
        if (is_callable(array($job, 'saveProcessSummaryCounts'))) {
          $job->saveProcessSummaryCounts($info['process_context']['drunkins_process_sub_id'], FALSE, $info['process_context']);
        }
      }
    }
  }

  if ($change) {
    $current_queue = $queue_name;
  }

  if ($current_queue) {
    // Get reference to full status info so the static array can be manipulated.
    $info = &drunkins_job_status_info($current_queue);

    if (isset($info['process_context']) && $info['running'] == 3 && $info['method'] !== 'batch') {
      // Add a new row for a 'sub process' (a part of the process that will end
      // when this function gets called again so we have a checkpoint to record
      // statistics).
      if ($change && !empty($info['process_context']['drunkins_process_id']) && variable_get('drunkins_process_summary_checkpoint', FALSE)) {
        if ($suppress_new_checkpoint) {
          // Make sure the next call to this function doesn't update the
          // previous checkpoint.
          unset($info['process_context']['drunkins_process_sub_id']);
        }
        else {
          $info['process_context']['drunkins_process_sub_id'] = db_insert('drunkins_process_run')
            ->fields(['pid' => $info['process_context']['drunkins_process_id'], 'name' => $current_queue, 'started' => time()])
            ->execute();
        }
      }

      return $info['process_context'];
    }
  }
  // Return (a 'dead end' reference into) an empty array if $current_queue info
  // does not exist, has no context, or the process is not running or not in
  // queue mode. This way, calls to this function for nonexistent jobs cannot
  // pollute 'job status info'. (Unlike calls to drunkins_job_status_info().)
  $return = array();
  return $return;
}

/**
 * Shutdown function: Writes the current context to a persistent cache.
 *
 * @see drunkins_current_context()
 */
function drunkins_flush_info() {
  // This function is only registered from drunkins_current_context() so we
  // know there actually is a 'current context', and we don't need to put
  // this function in the .module.

  // We can use this to detect fatal errors like Out Of Memory too. In that
  // sense it pairs with the mail function in drunkins_current_context(), which
  // would not be triggered if OOM was encountered while serializing context.
  // It's not water tight, though; fatal errors that happen when _we_ are
  // calling drunkins_current_context() will not be caught.
  $error = error_get_last();
  if ($error && $error['type'] ^ E_ERROR && ($mailto = variable_get('drunkins_fatal_error_mailto', ''))) {
    _drunkins_mail_fatal_error($mailto, "File: $error[file]\nLine: $error[line]\nMessage: $error[message]");
  }

  // Write current context.
  drunkins_current_context('');

  // Dereference class instances. (Probably unnecessary but doesn't hurt.)
  $info = &drunkins_job_status_info();
  $info = array();
}

/**
 * Send simple mail about fatal errors, bypassing hooks/alters.
 *
 * It's not sure whether this functionality will stay, in its current form. For
 * now it's dubbed "stopgap that seemed useful, but has not been tested much
 * in practice".
 */
function _drunkins_mail_fatal_error($to, $info) {
  // Pick the code from drupal_mail() that allows for configuration of the
  // mailer and has little chance of blowing up. Skip the alters and the
  // unneeded formatting.
  $default_from = variable_get('site_mail', ini_get('sendmail_from'));

  // This is what is left of drupal_mail() $message when you strip away the
  // stuff used by alters etc:
  $message = array(
    // 'from' is unecessary - except maybe for some obscure mailer class?
    'from'      => $default_from,
    'headers' => array(
      'From' => $default_from,
      'Sender' => $default_from,
      'Return-Path' => $default_from,
      'MIME-Version' => '1.0',
      'Content-Type' => 'text/plain; charset=UTF-8; format=flowed; delsp=yes',
      'Content-Transfer-Encoding' => '8Bit',
      'X-Mailer' => 'Drupal',
    ),
    'to'      => $to,
    'subject' => 'Drunkins encountered possibly-fatal error in ' . check_plain(variable_get('site_name', 'Drupal')),
    'body'    => "Info provided:\n\n$info",
  );

  $system = drupal_mail_system('drunkins', 'fatal_error');
  // DefaultMailSystem::format() does a LOT of proccessing, which can cause OOM
  // if $info is large. But since mail() also does at least a preg_replace(),
  // let's just do it. At least the contents will be safe. And if we encounter
  // an OOM it should be caught by the shutdown function which will then send a
  // much smaller mail.
  $message = $system->format($message);
  $system->mail($message);
}

/**
 * Sets or resets PHP memory limit.
 *
 * If we recorded that we already set the same (/ possibly a higher) limit with
 * a previous call, exit.
 *
 * @param string $job_id
 *  Job ID, used for logging. (Get rid of this parameter when writing runner.)
 * @param string $new_limit
 *   (Optional) limit to set. If not passed, an old limit is reset if that is
 *   known (and if it isn't known, the function exits without doing anything.)
 * @param bool $downsize
 *   (Optional) if TRUE, allow lowering of the current memory limit. If FALSE,
 *   only set if it is higher than the current limit. If $new_limit is empty,
 *   this parameter is ignored and lowering is allowed.
 *
 * @return int|bool
 *   If the limit could be set: the value of memory_limit before the call. FALSE
 *   if the new limit could not be set (because the given limit is lower than
 *   the current one, or because the ini_set() call failed).
 */
function _drunkins_set_memory_limit($job_id, $new_limit = '', $downsize = FALSE) {
  $initial_limit = &drupal_static(__FUNCTION__, 0);

  if (empty($initial_limit) && empty($new_limit)) {
    // We can't set an old initial value because we have not been called before.
    return FALSE;
  }

  if (empty($initial_limit) || !$downsize) {
    // Get current limit, to remember or to compare to $new_limit.
    $current_limit = parse_size(ini_get('memory_limit'));

    if (empty($initial_limit) ) {
      $initial_limit = $current_limit;
    }

    if ($new_limit) {
      if (!$downsize && parse_size($new_limit) <= $current_limit) {
        // Don't downsize.
        return FALSE;
      }
    }
    elseif ($initial_limit == $current_limit) {
      // Don't do ini_set if it's a no-op.
      return $current_limit;
    }
  }

  // Set new limit if provided; set initial limit otherwise.

  $old = @ini_set('memory_limit', $new_limit ? $new_limit : $initial_limit);

  if (!$old) {
    _drunkins_log($job_id, 'Failed setting memory limit to @mem', array('@mem' => $new_limit ? $new_limit : $initial_limit), WATCHDOG_WARNING);
  }

  return $old;
}

/**
 * Returns the PHP memory limit before any _drunkins_set_memory_limit() calls.
 *
 * @return int
 *   Initial memory limit; 0 if _drunkins_set_memory_limit() was never called.
 */
function _drunkins_get_initial_memory_limit() {
  return drupal_static('_drunkins_set_memory_limit', 0);
}

/**
 * Get configuration for currently running process.
 *
 * Internal function; doesn't have a lot of error checking.
 *
 * @param string $job_id
 * @param string $subsystem
 *   (optional) Subsystem / type of configuration:
 *   - 'runner' (default)
 *   - 'job'
 *   - 'items': returns the custom array of items to process in the job that is
 *     being started at the moment. FALSE if not set. $subset argument must be
 *     the default / not passed.
 * @param string $subset
 *   (optional) Subset of configuration:
 *   - 'running_process': options necessary for a running process. For subsystem
 *     'job', This may include a special key 'drunkins_start_items_overridden'
 *     which is set to TRUE if custom items were set for this job.
 *   - 'default': default configuration.
 *
 * @return array|bool
 */
function _drunkins_get_process_config($job_id, $subsystem = 'runner', $subset = 'running_process') {
  $info = drunkins_job_status_info($job_id);

  switch ("$subsystem.$subset") {
    case 'runner.running_process':
      // If 'process settings' not set yet, return the fixed configuration for
      // this job. (This makes it less necessary to guess whether one is
      // already allowed to get the "all" config / know when the custom config
      // is already set.)
      $config = !empty($info['process_runner_settings']) ? $info['process_runner_settings'] :
        (!empty($info['runner_settings']) ? $info['runner_settings'] : array());
      break;

    case 'runner.default':
      $config = !empty($info['runner_settings']) ? $info['runner_settings'] : array();
      break;

    case 'job.running_process':
      $config = !empty($info['process_settings']) ? $info['process_settings'] :
        (!empty($info['settings']) ? $info['settings'] : array());
      break;

    case 'job.default':
      $config = !empty($info['settings']) ? $info['settings'] : array();
      break;

    case 'items.running_process':
      $config = drupal_static('drunkins_set_process_items', FALSE);
      break;

    default:
      $config = FALSE;
  }

  return $config;
}

/**
 * Set configuration for currently running process.
 *
 * Configuration cannot be set incrementally with subsequent calls; if it is set
 * then previous 'dynamic' settings are erased. (They are just merged with any
 * 'static' configured settings.) This means we can run/finish a job and set up
 * another starting job with different settings, in the same call.
 *
 * When doing this for a batch process (after the process was already started),
 * it is likely that the settings are not persisted in the database, so be
 * careful. (It's not advised in the first place to call this function ever,
 * after the process was already started - but just saying. There's at least
 * one place which does this.)
 *
 * @param string $job_id
 * @param string $subsystem
 *   (optional) Subsystem to set configuration for:
 *   - 'runner' (default).
 *   - 'job'.
 *   - 'items': custom set of items to start this job with. (Hint:
 *     clearing them when not needed anymore, might free memory.) The 'job'
 *     subsystem must not be set anymore after 'items' is set; if that is done
 *     then a special flag signifying that custom items are set, is lost. (It's
 *     not hugely important, but still it's nice to have.)
 *
 * @param array $custom_settings
 *   Array of settings. in case $subsystem == 'items': an array of items.
 */
function _drunkins_set_process_config($job_id, $subsystem, array $custom_settings) {
  $info = &drunkins_job_status_info($job_id);

  switch ($subsystem) {
    case 'runner':
      if (isset($info['runner_settings']) && is_array($info['runner_settings'])) {
        $custom_settings += $info['runner_settings'];
      }
      $info['process_runner_settings'] = $custom_settings;
      break;

    case 'job':
      if (isset($info['settings']) && is_array($info['settings'])) {
        $custom_settings += $info['settings'];
      }
      $info['process_settings'] = $custom_settings;
      break;

    case 'items':
      $items = &drupal_static('drunkins_set_process_items');
      $items = $custom_settings;
      // Modify job settings to remember the items were overridden. This must
      // be done on set (not get) because otherwise this won't be saved in the
      // database. (We do not actually use this property anywhere but we might.)
      // This must be set after any 'job' setting, because modifying 'job' does
      // not keep this setting. (Which is on purpose; a new 'job' setting call
      // may pertain to a completely newly started process.)
      $info['process_settings']['drunkins_start_items_overridden'] = TRUE;
      break;
  }
}

/**
 * Alters queue info by adding info on processes.
 * Internal; meant only to be called from our hook_cron_queue_alter().
 *
 * @see drunkins_job_status_info()
 */
function _drunkins_queue_info_alter(array &$queues) {
  // Code note: drupal_alter('queue_info') may be called twice: independently,
  // and through drunkins_job_status_info(). (In D8: drupal_cron_run() does
  // this; in D7, only drunkins_overview().)

  $process_info = db_query("SELECT * FROM {drunkins_process}")->fetchAllAssoc('queue_name', PDO::FETCH_ASSOC);

  foreach ($queues as $queue_name => &$queue_info) {

    // Check data and merge 'process info' & calculated values. Only queues with
    // 'drunkins' set are important; others should never have 'process info'.
    if (!isset($queue_info['drunkins'])) {
      continue;
    }

    $info =& $queue_info['drunkins'];

    // Add 'process info' from database.

    if (isset($process_info[$queue_name])) {

      if ($process_info[$queue_name]['started'] > $process_info[$queue_name]['finished']) {
        // For running processes, merge only 'known' keys from the data column,
        // into 'process_*' properties:
        $process_data = unserialize($process_info[$queue_name]['data']);
        // If unserializing fails, _drunkins_cron_process_queue_item() will
        // discard all further queued items processed because context is empty.
        unset($process_info[$queue_name]['data']);

        if (isset($process_data['settings']) && is_array($process_data['settings'])) {
          $info['process_settings'] = $process_data['settings'];
        }
        if (isset($process_data['context']) && is_array($process_data['context'])) {
          $info['process_context'] = $process_data['context'];
        }
        if (isset($process_data['runner_settings']) && is_array($process_data['runner_settings'])) {
          $info['process_runner_settings'] = $process_data['runner_settings'];
        }
      }

      $info = array_merge($info, $process_info[$queue_name]);
    }

    // Calculate dynamic info.
    _drunkins_calculate_dynamic_info($info, $queue_name);
  }
  // For future re-coding, note $info / $queue_info still points inside $queues.
}

/**
 * Calculates dynamic values from other job info components; these are 'running'
 * plus optionally 'run_next' timestamp (which is only used by UI).
 *
 * @param array $info
 * @param string $job_id
 * @param bool $force_calc_next
 *   If TRUE, always calculate next cron-time if possible. By default it does
 *   not get calculated if the job is still running because it won't be checked
 *   by any process runner. Pass TRUE if you want to e.g. display it in a list.
 *
 * @see drunkins_job_status_info()
 */
function _drunkins_calculate_dynamic_info(array &$info, $job_id, $force_calc_next = FALSE) {
  // Just to be sure, protect against external sources setting this:
  unset($info['running']);
  unset($info['run_next']);

  if (!isset($info['run_info'])) {
    // No info (current/previous process run stored in database) available.
    // Calculate using only info from hook_queue_info().
    $info['running'] = isset($info['cron_schedule']) ? 1 : 0;
    return;
  }

  if ($info['run_info'] == 3) {
    // "error encountered while finishing process". (Only for advanced methods.)
    $info['running'] = 2;
  }
  elseif ($info['run_info'] == 1 || $info['run_info'] == 2 || $info['started'] > $info['finished']) {
    // s>f: Job is still running. (We're not verifying by checking queue items.)
    // 1: "running' or "running start; don't touch it".
    // 2: "running but start() should be rerun". (Only for advanced runners.)
    $info['running'] = 3;
  }

  if (isset($info['cron_schedule'])
      && (!isset($info['running']) || $force_calc_next)) {
    // Check time of the next cron run - either for calculating status or
    // because we've been told to.

    // If cron_schedule is zero/empty, the job should always be started, and
    // we can't calculate run_next.
    if (empty($info['started']) || empty($info['cron_schedule'])) {
      if (!isset($info['running'])) {
        $info['running'] = 1;
      }
      return;
    }

    // Check schedule against previous start time. (Not finish time; this way,
    // if e.g. an hourly job once runs for 61 minutes -and cron runs often- it
    // will still be restarted within the same hour, just after it finishes.)
    elseif (is_numeric($info['cron_schedule'])) {
      // Interval is minutes; check against 'last started' timestamp.
      // If cron is run every X minutes and $info['cron_schedule'] is a multiple
      // of that interval, we can miss a cron run if we compare the exact
      // same number of seconds, so substract some from $info['cron_schedule'].
      $info['run_next'] = $info['started'] + $info['cron_schedule'] * 60 - 10;
    }
    else {
      // Check whether this is a valid period string or crontab value. Since
      // these are complicated formats, just feed them through the
      // parsers and see if the return value is valid.
      $crontab = new DrunkinsCronTab($info['cron_schedule']);
      if ($crontab->cron) {
        $info['run_next'] = $crontab->nextTime($info['started']);
      }
      else {
        // date_modify will raise a PHP warning for invalid string. We can't use
        // date_add(..., date_interval_create_from_date_string($info['cron_schedule']))
        // instead; an 'empty' interval will be created without warning for
        // invalid string, but date_add() will yield a wrong date (php bug,
        // apparently) on PHP5.3. PHP5.5 is OK.
        $next = date_modify(date_create('@' . $info['started']), $info['cron_schedule']);
        if ($next) {
          $info['run_next'] = $next->getTimestamp();
        }
      }
      if (!isset($info['run_next'])) {
        watchdog('drunkins', 'Job "@job" cannot be scheduled: interval %i cannot be parsed.',
          array('@job' => $job_id, '%i' => $info['cron_schedule']), WATCHDOG_ERROR);
      }
    }
  }

  if (!isset($info['running'])) {
    $info['running'] = isset($info['run_next']) && REQUEST_TIME >= $info['run_next'] ? 1 : 0;
  }
}

/**
 * (Check whether we should) start/finish any jobs due/running on cron.
 *
 * We wanted to do this on hook_cron_queue_info (and until 2016 we did the
 * 'check_finish' part on cron), among others so that it can influence the
 * 'time' property of cron queues. (And possibly for other reasons, to decipher
 * from comments below.) But:
 * - We'd need a lock to prevent concurrent processes from working on the same
 *   job (i.e. concurrent start() calls - or immediate finish-after-start by a
 *   concurrent PHP process, which is the bug we saw in 2016) so until we have
 *   that, we can only executed from hook_cron which has its own locking.
 * - This would mean that every time someone checks the info (e.g. in the
 *   Elysia Cron UI screen) this gets executed. We probably really want to
 *   constrain this to cron runs.
 */
function _drunkins_check_cron_jobs($compiling_queue_info = FALSE) {
  // UPDATE 2020:
  // 1)
  // History: (re)running start() has always been done from drunkins_cron()
  // (either directly or, since June 2016, indirectly by calling this new
  // function). Running finish() was first done in drunkins_queue_info() and
  // since June 2016, through drunkins_cron() by this function. If we change
  // things around, we should likely keep in mind:
  // - concurrency, mentioned above
  // - not running on non-cron runs, mentioned above
  // - what could happen if we allow first running finish() and then start()
  //   for the same process in the same PHP request. Maybe the sea of comments
  //   below has some hints on that; I don't know/remember at the moment.
  // 2)
  // There's a sea of comments from June 2016 both here and in
  // drunkins_cron(_queue_info)() which I'm not going to touch now, just
  // highlight that
  // - a queue_info comment says "'process_items_after_rerun_start' DOES NOT
  //   WORK at the moment. A job will (with?) that set, will not start.".
  // - a former hook_cron comment (copied here and deleted there) said "items
  //   which are queued while starting a job, may be processed within the same
  //   HTTP request. @todo recheck the above statement later. While this is
  //   true now, I think we're aiming for a situation where this is not the
  //   case - probably only in a child runner class that is capable of
  //   rerunning start."
  // So, while the first point about the setting may be true, the practical
  // 'process_items_after_rerun_start' behavior IS currently true. (Which is
  // fine by me at the moment.)

  if ($compiling_queue_info) {
    // Because of the 2 points in the function comment, we could only do this
    // here if we are sure that we're being executed inside a cron process.
    // (Then we don't need to do funny hacks in order to incorporate the time
    // spent by this function into the 'time' property of cron queues.) But
    // that sounds too icky to do something about, now.
    return NULL;
/* This is what we documented and should still implement - ALONG with morphing
   'process_items_after_rerun_start' into a reversed capability as documented in
   @todos at drunkins_start_job().

'process_items_after_rerun_start' does not work anymore, at the moment; we don't
care because we do not have any processes that rerun start.

For now, the default is 'process_items_after_rerun_start == FALSE' which means
we don't have elysia2.1 problems by default and we can just return on $compiling_queue_info.
If we make it a reversed capability we will have elysia2.1 problems, so take
care of that. (Make a module-wide setting, or just check elysia version?)


I drafted the following desired behavior, when trying to understand things:

 * When to call what:* <-- UPDATE Dec 2020: see function comments; likely nothing except cron is applicable.

start / enqueue-items-restart:
 - if 'process_items_after_rerun_start' is not set for a job:
   on cron; return items normally in queue-info.
 - if 'process_items_after_rerun_start' is set AND NOT elysia2.1-compat:
   on cron; return items with 'skip on cron' in queue-info.
 - if 'process_items_after_rerun_start' is set AND elysia2.1-compat:
   on cron_queue_info; return queue-info depending on whether process_more is set.
     ^^ <== WE CANNOT do enqueue-items-restart then. If we do that, we might run it twice. Except if we make a new status 'starting' that we alternate between 'starting' and 'running' when running start() / waiting for the next cron run.
        (which is similar to the reason why we have commented out check-finish for now.)
@todo it probably makes sense to have 'process_items_after_rerun_start'=TRUE be
  the default (if this thing gets to be a 'capability'), but then simple processes
  running start() will suffer too. We should move all this into a child runner class.

check-finish:
 - if 'process_items_after_rerun_start' is not set for a job:
   on cron. (We can always just start processing items.)
   (We could maybe make a setting to do it on queue-info but I don't see why.)
 - if 'process_items_after_rerun_start' is set:
   - on cron_queue_info (because we don't know yet whether we can return queue-info)
   - OR on cron but prevent calling start() afterwards so that no queue items will
     be processed during this cron run
     ...depending on runner-ish setting 'prevent_cron_restart_delay'.

 * current situation: *

*for now*, rerunning start *probably* works with us doing everything on cron,
given that the default 'process_items_after_rerun_start' is FALSE at the moment.
DO NOT set it to true, because those jobs will not start. (drunkins_cron_queue_info()
does not return queue info for those processes, since our rewrite to accommodate
elysia 2.1.)


as long as you never set 'process_items_after_rerun_start' to FALSE.

@todo I only briefly checked but it seems that elysia_cron dev > 2.1
  calls this hook a lot. Do we want to cut down on the numberOfItems()
  calls in that case? Should we make a global option that ignores any
  'prevent_cron_restart_delay'? (Maybe even make that the default, and note at
  'prevent_cron_restart_delay' that you can only use this if you set a global variable?)
  ^^
  Dec 2020: I'm not going to change this now but: we changed things so this
  gets executed only once per HTTP request. (Even if we would change this to be
  executed on not-hook_cron - which we likely won't: see function comment.)
 */
  }

  // Even though it isn't applicable at this moment: accommodate for this
  // function being called multiple times per request. It should only do its
  // actions once - and return the time spent on it to every caller.
  $time_spent = &drupal_static(__FUNCTION__);
  if (isset($time_spent)) {
    return $time_spent;
  }
  $start_time = time();

  $reset_memory_limit = FALSE;

  // Set $job_info by reference because we need the up to date info if it gets
  // changed in _drunkins_queue_process_check_finish().
  foreach (drunkins_job_status_info() as $job_id => &$job_info) {

    if ($job_info['running'] == 3 && $job_info['method'] === 'queue') {
      if (empty($job_info['process_context'])) {
        // This might help debugging very strange conditions which have very
        // occasionally occurred. We'll just let it fall through to
        // _drunkins_queue_process_check_finish().
        $data = db_query('SELECT data FROM {drunkins_process} WHERE queue_name=:name', [':name' => $job_id])->fetchField();
        _drunkins_log($job_id, "Context is lost from status info; this must never happen. On disk, the data column is currently: @data", ['@data' => $data], WATCHDOG_CRITICAL);
      }

      if (!empty($job_info['process_context']['drunkins_process_more'])) {
        // The job's start() should be called again to fetch / enqueue more
        // items.
        // Please note: the only thing that prevents multiple start() calls
        // (which can take a long time) from overlapping in time, is the cron
        // lock. This module does not do any locking by itself.
//@todo solve this by alternating between a new 'starting' state (which only is set
// during start() call) and 'running' state (which is set while waiting for the next start() call)?
// If we don't do this, we certainly won't be able to call this from hook_cron_queue_info().
        _drunkins_enqueue_items($job_id);
      }
      else {
        // If a job is about to finish (i.e. has 0 items left) and can be
        // restarted on this same cron run, its queue-info must not have a
        // worker callback if the 'process_items_after_rerun_start' runner
        // setting is true. This is why, in that case,
        // - we must either call _drunkins_queue_process_check_finish() *before*
        //   determining the exact queue-info to return on hook_cron_queue_info,
        // - or we must call it with 'job_may_restart_immediately' set to FALSE.
        //
        // Doing the former means:
        // - we can still restart an already-finished job in the same cron run.
        //   (The latter introduces a delay of 1 cron run, so is not good for
        //   jobs that run very often - unless they have almost no chance of
        //   getting into an 'unfinished with 0 items to process' state.)
        // - we make superfluous Queue::numberOfItems() calls (to doublecheck if
        //   a process marked as 'running' still has items) on almost every
        //   hook_cron_queue_info() invocation.
        // - hook_cron_queue_info() could do a lot of processing (if a process
        //   actually gets finished).
        // - unfinished jobs with 0 items are finished also outside of cron runs
        //   when hook_cron_queue_info() is invoked, e.g. when someone views the
        //   job queue UI overview. (This might be a good thing; potentially
        //   confusing for a developer but less confusing for most users.)

        // @todo maybe now that we are clear what needs to happen (see temporary blurb at start of this function),
        //   the above comment is a bit much (because the former is not the default /
        //   is hardly ever done) and can be reworded/moved elsewhere?

        // (In some -dev version between 2015 and 20160625, we chose the former.
        // This introduced a bug, however - while a PHP process was running
        // start(), some other PHP process could execute hook_cron_queue_info()
        // (when viewing the drunkins UI) and 'finish' the starting process,
        // thereby breaking it.
        // So we moved it back to hook_cron(); the 'race condition' does not
        // exist there because Drupal takes care that only one PHP process runs
        // cron.

        //@todo probably, for 'process_items_after_rerun_start' (which will probably be the new default):
        // - add job_may_restart_immediately=FALSE
        // - check this in the start call just below, so we don't call it if we shouldn't.
        // (We don't need to do that now, because 'process_items_after_rerun_start'
        //  does not work / processes never get started. And the whole thing
        //  probably will get moved to a subclass. So we don't bother for now.)

        _drunkins_queue_process_check_finish($job_id, array('job_may_restart_immediately' => TRUE));
      }
    }

    if ($job_info['running'] == 1) {
      // No process is running but it should be.
      drunkins_start_job($job_id, 'queue');
    }
  }

  if ($reset_memory_limit) {
    // If any 'memory_limit_start' was (probably) set in a
    // _drunkins_enqueue_items() call and not reset, reset it here. If
    // it wasn't, this call does nothing / reapplies current limit.
    _drunkins_set_memory_limit('-');
  }

  // Set static variable.
  $time_spent = time() - $start_time;
  return $time_spent;
}
